#%%
import pandas as pd
from collections import Counter
from tqdm.auto import tqdm
import json
import os
import csv
from time import time # To time our operations
tqdm.pandas()
import re  # Import the re module
import numpy as np
import plotly.express as px
#%%
def clean_column(df, column):
    df[column] = df[column].str.lower()
    df[column] = df[column].fillna('')
    # Replacement operations
    df[column] = df[column].replace(r'\!', '', regex=True)
    df[column] = df[column].replace(r'[–—―‒]', '-', regex=True)  # Replace different dash characters with hyphen
    df[column] = df[column].replace(r'\.\.\.', ' ', regex=True)
    df[column] = df[column].replace(r'\*', ' ', regex=True)
    df[column] = df[column].replace(r'\/', ' ', regex=True)
    df[column] = df[column].replace(r'\°', ' ', regex=True)
    df[column] = df[column].replace(r'"', '', regex=True)
    df[column] = df[column].replace(r'‶', '', regex=True)
    df[column] = df[column].replace(r'«', '', regex=True)
    df[column] = df[column].replace(r'\.\-', '', regex=True)  # Remove all occurrences of '.-'
    df[column] = df[column].replace(r"^[-.',]", '', regex=True)  # Remove leading '-' or '.'
    df[column] = df[column].replace(r"^[-.',]", '', regex=True)  # Remove leading '-' or '.'
    df[column] = df[column].replace(r"^[-.',]", '', regex=True)  # Remove leading '-' or '.'
    df[column] = df[column].replace(r'�', '', regex=True)
    df[column] = df[column].replace(r'’', "'", regex=True)
    df[column] = df[column].replace(r'\[', " ", regex=True)
    df[column] = df[column].replace(r'\]', " ", regex=True)
    df[column] = df[column].replace(r'\(', " ", regex=True)
    df[column] = df[column].replace(r'\)', " ", regex=True)
    df[column] = df[column].replace(r'[ \t]{2,}', " ", regex=True)  # Collapse multiple spaces/tabs to a single space
    df[column] = df[column].str.strip()  # Remove leading and trailing spaces
    return df
#%%

#%%
def exact_ngram_match(df):
    filtered_df = df[df['gender'] == 'Not_Assigned']

    male_ngram_list = ['il', '-il', 'messieurs', 'mossieur', 'lui-même', 'le', 'celui-ci', 'le même', 'monsieur', 'grand-père', 'papa', 'celui', 'prêtre', 'homme', 'père', 'auquel', 'lequel', '-le', 'frère', 'm.', 'prêtres', 'garçon', 'un', 'mari', 'magicien', 'manants', 'matelots', 'monde', 'lesquels', 'praticien', 'pape', 'paysan', 'salop', 'sire', 'maître', 'seigneur', 'patron', 'général', 'prince', 'chevalier', 'commissaire', 'comte', 'chéri', 'commandant', 'augustin', 'ami', 'messire', 'duc', 'citoyen', 'petit', 'romain', 'baron', 'chef', 'mec', 'colonel', 'duquel', 'lieutenant', 'cousin', 'roi', 'bonhomme', 'jonas', 'cher', 'professeur', 'parrain', 'vicomte', 'berger', 'maréchal', 'brigadier', 'cadet', 'tonton', 'voleur', 'gendre', 'page', 'gamin', 'sir', 'assassin', 'cardinal', 'major', 'soldat', 'moine', 'fiston', 'curé', 'vieillard', 'brigand', 'coquin', 'cavalier', 'marin', 'gouverneur', 'malin', 'inspecteur', 'gringalet', 'messer', 'monsir', 'musicien', 'mort', 'voisin', 'président', 'évêque', 'oncle', 'député', 'cocher', 'aimé', 'compagnon', 'caporal', 'neveu', 'avocat', 'officier', 'connétable', 'signor', 'canonnier', 'pélerin', 'beau-père', 'parisien', 'étranger', 'prévot', 'malais', 'magistrat', 'lutin', 'grand-papa', 'amant', 'comtois', 'musard', 'barbier', 'marchand', 'sculpteur', 'négociant', 'croyant', 'midshipman', 'viennois', 'fils', 'frères', 'gars', 'garçons', 'seigneurs', 'guerrier', 'princes', 'rois', 'moines', 'sultan', 'pasteur', "l' un", 'celui - ci', 'celui -là', 'lui - même', 'les deux hommes', 'deux hommes', 'ces deux hommes', 'les trois hommes', 'trois hommes', 'des deux hommes', 'quatre hommes', 'plusieurs hommes', 'les quatre hommes', "-l' un", 'mon cher', 'du même', 'moi tout seul', 'le christ', 'lui aussi', 'mon père', 'son père']

    female_ngram_list = ['elle', 'elles', '-elle', 'elle-même', 'la', 'celle-ci', 'celle', 'la même', 'madame', 'maman', 'laquelle', 'femme', 'mademoiselle', 'femmes', 'toutes', "grand'mère", 'est-elle', 'chacune', 'auxquelles', 'filles', 'mesdames', 'quelle', 'mère', 'grand-mère', 'mesdemoiselles', 'vierges', 'épouses', 'reine', 'mères', 'salope', 'dame', 'chérie', 'nana', 'angélique', 'miss', 'princesse', 'manette', 'une', 'comtesse', 'maîtresse', 'fillette', 'chère', 'mignonne', 'reinette', 'duchesse', 'nounou', 'cocotte', 'citoyenne', 'cousine', 'tante', 'nourrice', 'marraine', 'demoiselle', 'cette', 'baronne', 'bonne-maman', 'amie', 'mémé', 'fille', 'chrétienne', 'sœur', 'mamzelle', 'malheureuse', 'madone', 'bohémienne', 'bergère', 'mamselle', 'friponne', 'égyptienne', 'aïeule', 'celles', 'lesquelles', 'dames', 'mamie', 'ele', "l' une", 'celle - ci', 'celle -là', 'elle - même', 'nous toutes', "la sienne", 'moi toute seule', 'la tienne', 'la bonne', 'elle aussi', 'la mienne', 'toute seule'
                        ]
    ambiguous_ngram_list = ['je', 'vous', 'ma', 'ta', 'sa', 'son', 'mon', 'ton', 'qui', 'dont', 'les', 'me', '-vous', '-je', 'ce', '-ce', '-nous', 'nous', 'ses', 'moi', 'moi-même', 'vous-même', 'tu', 'te', 'toi', 'toi-même', 'lui', "j'", "l'", 'ils', "m'", 'leur', 'on', 'mes', 'où', 'les', 'notre', 'leurs', "c'", 'votre', 'eux', 'nos', "t'", '-moi', "qu'", 'que', 'en', 'tous', 'vos', 'se', 'chacun', "s'", 'tes', '-ils', 'mienne', 'tienne', 'ceux', 'enfant', 'y', 'soi', 'sien', '-on', 'ça', 'docteur', 'gens', 'enfants', 'petite', '-toi', 'capitaine', 'ensemble', 'est-on', 'famille', 'philosophe', 'musiciens', 'miennes', 'miens', 'fit', 'fous', 'fut', '-lui', 'vôtre', 'témoins', 'amis', 'couple', 'mien', 'misérable', 'maire', 'violon', 'virtuose', 'violoniste', 'vôtres', 'écuyer', 'été', 'était', 'équipages', 'turcs', 'oiseau', 'oiseaux', 'poète', 'politiques', 'peuple', 'philosophes', 'plusieurs', 'quatre-vingt-dix-neuf', 'danseurs', 'siens', 'sentinelle', 'serait', 'personne', 'hommes', 'j', 'aubert', 'excellence', 'ange', 'camarade', 'tien', 'aux', 'imbécile', 'du', 'médecin', 'chien', 'bébé', 'juge', 'chouchou', 'pauvre', 'de', 'huguet', 'ile', 'vingt', 'casse-cou', 'tours', 'pompon', 'espagnol', 'allemand', 'corsaire', '-tu', 'canaille', 'bourreau', 'gui', 'li', 'ministre', 'peste', 'caro', 'animal', 'môme', 'flic', 'démon', 'moitié', 'vo', 'z', 'baba', 'bichon', 'gendarme', 'morbleu', 'guide', 'mé',  'jemmy', 'malade',  'majesté', 'politique', 'ermite', 'auxquels', 'i', 'autres', 'marquis', 'honoré', 'you', 'deux', 'certains', 'mike', 'treize', 'trois', 'paysans', 'époux', 'parents', 'spectateurs', 'personnes', "-j'", 'étudiants', 'voyageurs', 'voleurs', 'groupe', 'serviteurs', 'gendarmes', 'nôtres', "quelqu' un", 'moi - même', "d' autres", 'toi - même', 'vous - même', 'eux - mêmes', 'tous deux', 'tous les deux', 'les uns', 'ceux - ci', 'voyageurs', 'morts', 'amis', 'domestiques', 'nous deux', 'tous trois', 'nous - mêmes', 'nous autres', 'nous tous', 'nous trois', 'nous - même', 'nous quatre', 'nous cinq', 'nous seuls', 'nous mêmes', 'nous tout seuls', 'nous aussi', 'nous deux seuls', 'nous six', 'nous tous ensemble', 'nous autres poètes', 'toutes deux', 'toutes les deux', 'toutes trois', 'toutes les trois', 'toutes les quatre', 'toutes quatre', 'toutes ensemble', 'toutes les mêmes', 'tous les trois', 'quelques - uns', 'soi - même', 'messieurs - dames', 'ces messieurs - dames', 'messieurs dames', 'ces messieurs dames', 'les messieurs - dames', 'messieurs - mesdames', 'quelques messieurs - mesdames', 'des messieurs - dames', 'aux messieurs - dames', 'ces messieurs , dames', 'les hommes', 'vous autres', 'tous les hommes', 'vous deux', 'sienne', "l' autre", 'célébrité'
]


    singular_rows = filtered_df[(filtered_df['text'].isin(male_ngram_list))]
    df.loc[singular_rows.index, 'gender'] = 'Male'
    plural_rows = filtered_df[(filtered_df['text'].isin(female_ngram_list))]
    df.loc[plural_rows.index, 'gender'] = 'Female'
    ambiguous_rows = filtered_df[(filtered_df['text'].isin(ambiguous_ngram_list))]
    df.loc[ambiguous_rows.index, 'gender'] = 'Ambiguous'

    del filtered_df
    return df
#%%
def always_gender_head(df):
    filtered_df = df[df['gender'] == 'Not_Assigned']

    always_singular_male_head = ['père', 'mossieur', 'mr.', 'monsignor', 'défunt', 'connétable', 'homme', 'roi', 'mari', 'fils', 'ami', 'enquêteur', 'comte', 'frère', 'duc', 'abbé', 'empereur', 'colonel', 'prince', 'marquis', 'baron', 'vieillard', 'chevalier', 'maître', 'oncle', 'général', 'curé', 'officier', 'petit', 'inconnu', 'compagnon', 'vieux', 'patron', 'cardinal', 'garçon', 'vicomte', 'amant', 'commandant', 'président', 'cocher', 'assassin', 'étranger', 'cousin', 'chef', 'ingénieur', 'ennemi', 'bonhomme', 'prêtre', 'dernier', 'cher', 'malheureux', 'maréchal', 'moine', 'neveu', 'premier', 'lieutenant', 'banquier', 'anglais', 'avocat', 'directeur', 'amiral', 'prisonnier', 'major', 'inspecteur', 'peintre', 'gouverneur', 'brigadier', 'monsieur', 'soldat', 'grand-père', 'blessé', 'magistrat', 'évêque', 'type', 'agent', 'voyageur', 'chasseur', 'époux', 'indien', 'valet', 'messieurs', 'garçons', 'frères', 'fiancé', 'chauffeur', 'paysan', 'fou', 'huissier', 'interlocuteur', 'gardien', 'américain', 'intendant', 'cavalier', 'pape', 'policier', 'portier', 'seigneur', 'gamin', 'fermier', 'gentilhomme', 'gars', 'conducteur', 'marchand', 'canadien', 'jardinier', 'bossu', 'conducteur', 'mort', 'facteur', 'meurtrier', 'geôlier', 'greffier', 'parisien', 'français', 'géant', 'condamné', 'meunier', 'pêcheur', 'nain', 'mendiant', 'tailleur', 'messager', 'juif', 'page', 'caissier', 'pasteur', 'voleur', 'régisseur', 'pharmacien', 'dauphin', 'braconnier', 'forgeron', 'boucher', 'mécanicien', 'sous-préfet', 'fugitif', 'héros', 'clerc', 'caporal', 'chinois', 'berger', 'commandeur', 'sous-officier', 'pâtissier', 'tsar', 'sacristain', 'patient', 'moribond', 'chéri', 'espagnol', 'gendre', 'beau-père', 'chien', 'saint', 'ouvrier', 'passant', 'client', 'dieu', 'innocent', 'employé', 'laquais', 'savant', 'amoureux', 'matelot', 'marin', 'chrétien', 'criminel', 'écrivain', 'espion', 'mourant', 'gaillard', 'serviteur', 'voisin', 'adolescent', 'copain', 'bourgeois', 'allemand', 'écolier', 'visiteur', 'chirurgien', 'insensé', 'vivant', 'mec', 'malfaiteur', 'sorcier', 'coquin', 'italien', 'gentleman', 'protecteur', 'étudiant', 'ambassadeur', 'député', 'acteur', 'imposteur', 'musicien', 'rival', 'guerrier', 'ingrat', 'vagabond', 'scélérat', 'infirmier', 'sot', 'loup', 'souverain', 'second', 'revenant', 'aventurier', 'boulanger', 'comédien', 'européen', 'serrurier', 'sergent', 'beau-frère', 'saint-père', 'beau-frère', 'arrière-grand-père', "grand'père", 'arrière-arrière-grand-père', 'monseigneur', 'granpère', 'pseudo-père', 'accusé', 'préfet', 'aîné', 'homme-là', 'bourreau', 'hôtelier', 'instituteur', 'petit-fils', 'baronnet', 'archevêque', 'lion', 'chat', 'parrain', 'm.', 'papa', 'tuteur', 'sauveur', 'orateur', 'supérieur', 'prieur', 'conseiller', 'sieur', 'chanoine', 'cheval', 'lord', 'sculpteur', 'postillon', 'avoué', 'négociant', 'grand-duc', 'cuisinier', 'bienfaiteur', 'héritier', 'bien-aimé', 'prélat', 'châtelain', 'infortuné', 'tueur', 'sénateur', 'métayer', 'apprenti', 'cadet', 'religieux', 'vainqueur', 'grand-oncle', 'régent', 'venu', 'nègre', 'celui', 'colosse', 'bosseman', 'écuyer', 'chancelier', 'coiffeur', 'irlandais', 'écossais', 'consul', 'orphelin', 'armateur', 'inventeur', 'conteur', 'major', 'un', 'charretier', 'prévôt', 'routier', 'gros', 'bon', 'métis', 'grand', 'seul', 'breton', 'épicier', 'forçat', 'révérend', 'brigand', 'usurier', 'protégé', 'vicaire', 'cabaretier', 'chapelain', 'prévenu', 'barman', 'observateur', 'colaborateur', 'bûcheron', 'satan', 'saint-louis', 'saint-maur', 'saint-simon', 'espagnol', 'pédant', 'soudard', 'rejeton', 'hobereau', 'fluet', 'fripier', 'descendant', 'galant', 'vaillant', 'majordome', 'dormeur', 'trépassé', 'doge', 'tentateur', 'adjoint', 'substitut', 'serin', 'adorateur', 'prétendu', 'promeneur', 'affligé', 'écervelé', 'original', 'paroissien', 'feu', 'factionnaire', 'ancien', 'sémillant', 'singe', 'tigre', 'compère', 'canut', 'parvenu', 'dandy', 'financier', 'malicieux', 'collaborateur', 'farceur', 'absorbé', 'alsacien', 'impertinent', 'valet-de-chambre', 'penseur', 'futur', 'chapelier', 'bottier', 'caudataire', 'musico', 'chanteur', 'banqueroutier', 'danseur', 'bailli', 'praticien', 'sultan', 'correcteur', 'commensal', 'délaissé', 'étourdi', 'capitano', 'marié', 'collégien', 'arrivant', 'créateur', 'dénicheur', 'dévot', 'subordonné', 'jésus', 'laboureur', 'gentil', "grand'papa", 'limonadier', 'obligeant', 'particulier', 'voiturier', 'genevois',  'arménien', 'arrivé', 'bouffon', 'brutal', 'bêta', 'canari', 'canard', 'converti', 'dompté', 'fugitif', 'maladroit', 'minet', 'rédempteur', 'sous-lieutenant', 'turbulent', 'vieux-là'
]
    singular_male_rows = filtered_df[(filtered_df['mention_len'] <= 4)
                            & (filtered_df['number'].isin(['Singular']))
                            & (filtered_df['head_word'].isin(always_singular_male_head))]
    df.loc[singular_male_rows.index, 'gender'] = 'Male'

    always_female_head = ['celle', 'milady','donna', 'duègne', 'mairesse', 'mère', 'soeur', 'femme', 'auteure', 'reine', 'senora', 'fille', 'amie', 'comtesse', 'duchesse', 'princesse', 'maîtresse', 'sœur', 'marquise', 'baronne', 'tante', 'petite', 'inconnue', 'dame', 'veuve', 'cousine', 'fillette', 'nièce', 'servante', 'cousine', 'chère', 'vieille', 'chérie', 'compagne', 'orpheline', 'bonne', 'femmes', 'dames', 'filles', 'malheureuse', 'belle-mère', 'fiancée', 'grand-mère', 'belle-sœur', 'maman', 'petite-fille', 'soeur', "grand'mère", 'belle-fille', 'voisine', 'marraine', 'nourrice', 'gouvernante', 'rivale', 'filleule', 'protégée', 'protectrice', 'patronne', 'bienfaitrice', 'parente', 'cliente', 'belle', 'bien-aimée', 'visiteuse', 'poupée', 'copine', 'prisonnière', 'suivante', 'corsetière', 'coureuse', 'contrôleuse', 'contemplatrice', 'conscrite', 'consoeur', 'compagnonne', 'camériste', 'cadette', 'cuisinière', 'souveraine', 'favorite', 'confidente', 'couturière', 'grand-tante', 'demoiselle', 'danseuse', 'logeuse', 'petite-nièce', 'jumelle', 'conductrice', 'ménagère', 'colaboratrice', 'femelle', 'libératrice', 'captive', 'frangine', 'patiente', 'maitresse', 'gardienne', 'bergère', 'bourgeoise', 'passagère', 'concubine', 'promise', 'demi-sœur', 'muse', 'préférée', 'soubrette', 'supérieure', 'tutrice', 'collaboratrice', 'négresse', 'remplaçante', 'geôlière', 'étrangère', 'morte', 'folle', 'sainte', 'fée', 'religieuse', 'infirmière', 'paysanne', 'actrice', 'française', 'gamine', 'sorcière', 'épouse', 'vierge', 'ouvrière', 'voleuse', 'parisienne', 'institutrice', 'courtisane', 'anglaise', 'seule', 'déesse', 'comédienne', 'coquette', 'chrétienne', 'prostituée', 'mourante', 'criminelle', 'italienne', 'mendiante', 'innocente', 'allemande', 'madone', 'adolescente', 'sirène', 'pute', 'américaine', 'nymphe', 'ennemie', 'héritière', 'amoureuse', 'ingrate', 'pauvresse', 'amante', 'juive', 'martyre', 'coquine', 'bohémienne', 'blonde', 'aventurière', 'vivante', 'serveuse', 'héroïne', 'impératrice', 'brune', 'marchande', 'étudiante', 'aïeule', 'amazone', 'intrigante', 'espagnole', 'indienne', 'chanteuse', 'nonne', 'blanchisseuse', 'pécheresse', 'châtelaine', 'grande', 'gonzesse', 'cantatrice', 'vendeuse', 'insensée', 'chienne', 'vache', 'chinoise', 'gitane', 'lionne', 'gaillarde', 'lady', 'mariée', 'petites', 'amies', 'sœurs', 'demoiselles', 'belles', 'soeurs', 'tantes', 'auditrices', 'lectrices', 'mères', 'innocentes', 'jument', 'chatte', 'nounou', 'correspondante', 'belle-soeur', 'marâtre', 'lectrice', 'mignonne', 'coiffeuse', 'conjointe', "grand'tante", 'poule', 'consolatrice', 'pauline', 'mégère', 'greffière', 'camérière', 'caissière', 'chambrière', 'cavalière', 'première', 'mercière', 'meurtrière', 'lavandière', 'régulière', 'banquière', 'barbière', 'nourricière', 'particulière', 'sous-caissière', 'chevalière', 'chevrière', 'coéquipière', 'crémière', 'fermière', 'persécutrice', 'coadjutrice', 'directrice', 'génitrice', 'fondatrice', 'débitrice', 'dominatrice', 'dénonciatrice', 'détentrice', 'calomniatrice', 'créatrice', 'modératrice', 'rédactrice', 'spectatrice', 'présentatrice', 'interlocutrice', 'inspectrice', 'ambassadrice', 'traductrice', 'narratrice', 'ex-institutrice', 'réalisatrice', 'donatrice', 'opératrice', 'observatrice', 'alice', 'testatrice', 'tentatrice', 'éducatrice', 'éditrice', 'animatrice', 'organisatrice', 'triomphatrice', 'accusatrice', 'accompagnatrice', 'usurpatrice', 'admiratrice', 'factrice', 'enquêtrice', 'perceptrice', 'séductrice', 'navigatrice', 'simulatrice', 'monitrice', 'manipulatrice', 'auditrice', 'prestidigitatrice', 'sculptrice', 'répétitrice', 'prédatrice', 'préparatrice', 'conjuratrice', 'affabulatrice', 'incitatrice', 'inspiratrice', 'indicatrice', 'novatrice', 'délatrice', 'eurydice', 'employée', 'invitée', 'infortunée', 'possédée', 'habituée', 'pestiférée', 'exilée', 'noyée', 'condamnée', 'alliée', 'désespérée', 'naufragée', 'ressuscitée', 'traînée', 'accusée', 'damnée', 'divorcée', 'enragée', 'abandonnée', 'salariée', 'mijaurée', 'réprouvée', 'dévergondée', 'croisée', 'effrontée', 'droguée', 'déclassée', 'blessée', 'privilégiée', 'accouchée', 'épousée', 'passionnée', 'isolée', 'névrosée', 'brûlée', 'toquée', 'suicidée', 'réfugiée', 'émigrée', 'évaporée', 'inculpée', 'illuminée', 'handicapée', 'associée', 'agitée', 'affamée', 'aînée', 'entêtée', 'exaltée', 'défroquée', 'suppliciée', 'désœuvrée', 'écervelée', 'cinglée', 'adorée', 'athée', 'détraquée', 'dulcinée', 'délurée', 'demeurée', 'opprimée', 'passante', 'voyante', 'méchante', 'débutante', 'savante', 'surveillante', 'géante', 'convalescente', 'figurante', 'commerçante', 'vaillante', 'assistante', 'absente', 'infante', 'protestante', 'revenante', 'pénitente', 'insolente', 'indifférente', 'ignorante', 'élégante', 'suppliante', 'agonisante', 'croyante', 'charmante', 'postulante', 'présidente', 'démente', 'impertinente', 'indépendante', 'sympathisante', 'feignante', 'estivante', 'extravagante', 'gisante', 'gérante', 'arrivante', 'adjointe', 'enseignante', 'militante', 'imprudente', 'bacchante', 'soignante', 'fainéante', 'impudente', 'descendante', 'combattante', 'communiante', 'hôtesse', 'tigresse', 'enchanteresse', 'abbesse', 'chanoinesse', 'petite-maîtresse', 'archiduchesse', 'prêtresse', 'sous-maîtresse', 'drôlesse', 'diablesse', 'mulâtresse', 'doctoresse', 'ogresse', 'vicomtesse', 'devineresse', 'prophétesse', 'ivrognesse', 'poétesse', 'traîtresse', 'sauvagesse', 'chasseresse', 'druidesse', 'grande-duchesse', 'quakeresse', 'scélératesse', 'papesse', 'turquesse', 'stewardesse', 'vieillesse', 'ânesse', 'jésuitesse', 'peintresse', 'demi-déesse', 'contremaîtresse', 'docteresse', 'dernière', 'madame', 'gueuse', 'garce', 'salope', 'conne', 'souris', 'petite-là', 'marie', 'madeleine', 'vipère', 'idiote', 'catin', 'pimbêche', 'intruse', 'meuf', 'donzelle', 'orgueilleuse', 'vieille-là', 'françaises', 'saintes', 'compagnes', 'maréchale', 'louve', 'mademoiselle', 'douairière', 'meunière', 'mlle', 'mme', 'magicienne', 'une', 'venue', 'agée', 'grosse', 'métisse', 'seconde', 'protégée', 'défunte', 'voyageuse', 'dompteuse', 'ouvreuse', 'dormeuse', 'revendeuse', 'causeuse', 'empoisonneuse', 'procureuse', 'receveuse', 'goualeuse', 'rapporteuse', 'promeneuse', 'plaideuse', 'charmeuse', 'quêteuse', 'repasseuse', 'habilleuse', 'fileuse', 'brûleuse', 'brodeuse', 'placeuse', 'laveuse', 'baigneuse', 'maquilleuse', 'menteuse', 'rabouilleuse', 'boiteuse', 'curieuse', 'masseuse', 'chroniqueuse', 'magnétiseuse', 'tueuse', 'travailleuse', 'nageuse', 'paresseuse', 'régisseuse', 'balayeuse', 'solliciteuse', 'chercheuse', 'brocanteuse', 'bienheureuse', 'nourrisseuse', 'tricoteuse', 'teuse', 'conteuse', 'réveilleuse', 'pleureuse', 'boiteuse', 'parfumeuse', 'loueuse', 'tapeuse', 'rêveuse', 'fumeuse', 'pêcheuse', 'fosseuse', 'tailleuse', 'sainte-marie', 'espagnole', 'miss', 'pédante', 'donna', 'gitana', 'rejetonne', 'hoberelle', 'fluette', 'fripière', 'descendante', 'galante', 'vaillante', 'goule', 'intendante', 'gouvernante', 'dormeuse', 'trépassée', 'tentatrice', 'provinciale', 'recluse', 'lingère', 'adjointe', 'substitute', 'serine', 'adoratrice', 'prétendue', 'promeneuse', 'affligée', 'écervelée', 'originale', 'matrone', 'paroissienne', 'feue', 'ancienne', 'sémillante', 'canuse', 'parvenue', 'malicieuse', 'collaboratrice', 'absorbée', 'alsacienne', 'impertinente', 'future', 'chapelière', 'bottière', 'chanteuse', 'banqueroutière', 'danseuse', 'harpie', 'baillie', 'praticienne', 'sultane', 'correctrice', 'commensale', 'délaissée', 'étourdie', 'mariée', 'collégienne', 'arrivante', 'créatrice', 'dénicheuse', 'dévote', 'dauphine', 'subordonnée', 'maraîchère', 'laboureuse', 'gentille', "grand'maman", 'limonadière', 'obligeante', 'particulière', 'voiturière', 'genevoise', 'arménienne', 'arrivée', 'brutale', 'bêtasse', 'convertie', 'domptée', 'fugitive', 'maladroite', 'minette', 'née', 'rédemptrice', 'turbulente', 'écolière'

]
    female_rows = filtered_df[(filtered_df['mention_len'] <= 4)
                            & (filtered_df['number'].isin(['Singular']))
                            & (filtered_df['head_word'].isin(always_female_head))]
    df.loc[female_rows.index, 'gender'] = 'Female'


    always_ambiguous_head = ['gendarmerie', 'cortége', 'hôte', 'sainteté', 'seigneurie', 'chœur', 'altesse', 'ange', 'armée', 'assemblée', 'assistance', 'auditoire', 'bande', 'bataillon', 'bateau', 'belle-famille', 'bourg', 'brigade', 'brute', 'bébé', 'bête', 'cadavre', 'dépouille', 'camp', 'canaille', 'capitale', 'caravane', 'cavalerie', 'centaine', 'cinquantaine', 'cité', 'clergé', 'clientèle', 'cohorte', 'colonie', 'communauté', 'commune', 'commère', 'compagnie', 'comté', 'confrérie', 'connaissance', 'conseil', 'contrée', 'corporation', 'corps', 'cortège', 'couple', 'cour', 'couvent', 'couvée', 'crapule', 'créature', 'demi-douzaine', 'diocèse', 'divinité', 'division', 'dizaine', 'douzaine', 'démon', 'département', 'entourage', 'escadron', 'escorte', 'famille', 'fantôme', 'foule', 'garnison', 'garçonnière', 'gens', 'gentilhommière', 'gouvernement', 'groupe', 'horde', 'idole', 'individu', 'jeunesse', 'justice', 'milice', 'ministère', 'monastère', 'monde', 'monstre', 'multitude', 'nation', 'noblesse', 'nombreuses', 'nombreux', 'ombre', 'orchestre', 'ordre', 'parent', 'paroisse', 'patrouille', 'pays', 'personnage', 'personne', 'personnel', 'peuple', 'plupart', 'police', 'populace', 'population', 'province', 'public', 'quartier', 'quatrième', 'quinzaine', 'race', 'royaume', 'régiment', 'république', 'secteur', 'section', 'sentinelle', 'silhouette', 'société', 'spectre', 'star', 'statue', 'tablée', 'tribu', 'troupe', 'vermine', 'village', 'ville', 'vingtaine', 'âme', 'équipage', 'être', 'autres', 'personnes', 'habitants', 'compatriotes', 'enfants', 'compagnons', 'amis', 'soldats', 'parents', 'voyageurs', 'gens', 'monde', 'morts', 'domestiques', 'paysans', 'camarades', 'gendarmes', 'passants', 'passants', 'ouvriers', 'médecins', 'prisonniers', 'indiens', 'hôtes', 'allemands', 'officiers', 'pauvres', 'prussiens', 'ennemis', 'matelots', 'chiens', 'siens', 'convives', 'étrangers', 'flics', 'juges', 'gardes', 'assistants', 'invités', 'voisins', 'chasseurs', 'russes', 'clients', 'collègues', 'agents', 'maîtres', 'malades', 'bandits', 'vivants', 'voleurs', 'blessés', 'animal', 'ménage', 'majesté', 'victime', 'moi', 'lecteur', 'oiseau', 'amour', 'ours', 'personnages', 'colons', 'jumeaux'
]
    always_ambiguous_rows = filtered_df[(filtered_df['mention_len'] <= 4)
                                        & (filtered_df['head_word'].isin(always_ambiguous_head))]
    df.loc[always_ambiguous_rows.index, 'gender'] = 'Ambiguous'

    del filtered_df
    return df
#%%
def singular_ambiguous_gender_head(df):
    filtered_df = df.copy()

    # heads that can either be male or female. The context is used to disambiguate
    singular_ambiguous_head = ['enfant', 'reporter', 'autre', 'docteur', 'détective', 'capitaine', 'complice', 'médecin', 'commissaire', 'notaire', 'juge', 'artiste', 'domestique', 'malade', 'maire', 'concierge', 'ministre', 'professeur', 'auteur', 'esclave', 'troisième', 'coupable', 'misérable', 'pensionnaire', 'élève', 'somnambule', 'garde-malade', 'garde', 'secrétaire', 'modiste', 'camarade', 'chèvre', 'gosse', 'aveugle', 'aristocrate', 'gendarme', 'poète', 'commissionnaire', 'traître', 'lâche', 'guide', 'flic', 'témoin', 'philosophe', 'imbécile', 'sauvage', 'bandit', 'ivrogne', 'journaliste', 'militaire', 'pauvre', 'novice', 'russe', 'collègue', 'fleuriste', 'aubergiste', 'adversaire', 'propriétaire', 'môme', 'architecte', 'pilote', 'ermite', 'pote', 'interne', 'suisse', 'procureur', 'infirme', 'jeune', 'prophète', 'brave', 'drôle', 'monarque', 'mousquetaire', 'vétérinaire', 'poëte', 'photographe', 'apothicaire', 'matamore', 'apôtre', 'séminariste', 'commis', 'noble', 'drapier', 'fat', 'diable', 'diplomate', 'droguiste', 'paralytique', 'journalier', 'aide-de-camp', 'partenaire', 'dentiste', 'confrère', 'copiste', 'centenaire', 'timide', 'vaguemestre', 'patriote', 'convive', 'ex-jésuite', 'gnangnan', 'anatomiste', 'dr', 'fidèle', 'myope', 'proviseur'
                               ]

    vowel_singular_ambiguous_head = [head for head in singular_ambiguous_head if head.startswith(('a', 'e', 'é', 'è', 'ê', 'i', 'o', 'u', 'y', 'h'))]
    consonant_singular_ambiguous_head = [head for head in singular_ambiguous_head if head not in vowel_singular_ambiguous_head]

    male_vowel_prefix = ['au ', 'du ',  'un ', 'cet ', 'ce ', 'le ', 'cher', 'petit', 'premier', 'second', 'dernier', 'bel', 'ancien', 'malheureux', 'grand', 'intéressant', 'nouvel', 'quel', 'audacieux', 'heureux', 'infortuné ', 'le très']
    male_vowel_prefix += [prefix + ' ' + adj + ' ' for prefix in ['mon', 'ton', 'son', "l'", 'quel'] for adj in ['cher', 'petit', 'premier', 'second', 'troisième', 'dernier', 'pauvre', 'propre', 'brave', 'bel', 'ancien', 'jeune', 'malheureux', 'redoutable', 'grand', 'intéressant', 'nouvel', 'terrible', 'seul', 'excellent', 'fidèle']]
    male_vowel_prefix += [prefix + ' ' + adj + ' ' for prefix in ['notre', 'votre', 'leur', 'pauvre'] for adj in ['cher', 'petit', 'premier', 'second', 'dernier', 'bel', 'ancien', 'malheureux', 'grand', 'intéressant', 'nouvel', 'seul', 'excellent']]

    female_vowel_prefix = ['la ', 'cette', 'une', 'ma ', 'sa ', 'chère', 'petite', 'première', 'seconde', 'dernière', 'belle', 'ancienne', 'malheureuse', 'grande', 'innocente', 'intéressante', 'nouvelle', 'seule', 'excellente']
    female_vowel_prefix += [prefix + ' ' + adj + ' ' for prefix in ['notre', 'votre', 'leur', 'pauvre', "l'", 'quelle'] for adj in ['chère', 'petite', 'première', 'seconde', 'dernière', 'belle', 'ancienne', 'malheureuse', 'grande', 'innocente', 'intéressante', 'nouvelle', 'seule', 'excellente']]

    male_rows = filtered_df[(filtered_df['mention_len'] <= 4)
                            & (filtered_df['number'].isin(['Singular']))
                            & (filtered_df['gender'].isin(['Not_Assigned']))
                            & (filtered_df['text'].str.startswith(tuple(male_vowel_prefix)))
                            & (filtered_df['head_word'].isin(vowel_singular_ambiguous_head))]
    filtered_df.loc[male_rows.index, 'gender'] = 'Male'

    female_rows = filtered_df[(filtered_df['mention_len'] <= 4)
                            & (filtered_df['gender'].isin(['Not_Assigned']))
                            & (filtered_df['number'].isin(['Singular']))
                            & (filtered_df['text'].str.startswith(tuple(female_vowel_prefix)))
                            & (filtered_df['head_word'].isin(vowel_singular_ambiguous_head))]
    filtered_df.loc[female_rows.index, 'gender'] = 'Female'

    # -------consonant--ambiguous--head--------------------------------

    male_consonant_prefix = ['le ', 'du ', 'au ', 'un ', 'ce ', 'cet ', 'mon ', 'messire', 'ton ', 'son ', 'cher ', "l' ancien ", 'm. ', 'mm .', 'mm ', 'monsieur',  'quel ', 'cher ', 'petit ', 'premier ', 'second ', 'dernier ', 'bel ', 'ancien ', 'malheureux', 'grand ', 'intéressant ', 'nouveau', 'seul ', 'excellent ', 'bon ', 'audacieux ', 'heureux ', 'infortuné ', 'vieux ']
    male_consonant_prefix += [prefix + ' ' + adj + ' ' for prefix in ['notre', 'votre', 'leur', 'pauvre', "l'"] for adj in ['cher', 'petit', 'premier', 'second', 'dernier', 'bel', 'ancien', 'malheureux', 'grand', 'intéressant', 'nouvel', 'seul', 'excellent', 'bon', 'audacieux', 'heureux', 'infortuné', 'vieux']]
    male_rows = filtered_df[(filtered_df['mention_len'] <= 4)
                            &(filtered_df['mention_len'] > 1)
                            & (filtered_df['number'].isin(['Singular']))
                            & (filtered_df['gender'].isin(['Not_Assigned']))
                            & (filtered_df['text'].str.startswith(tuple(male_consonant_prefix)))
                            & (filtered_df['head_word'].isin(consonant_singular_ambiguous_head))]
    filtered_df.loc[male_rows.index, 'gender'] = 'Male'

    female_consonant_prefix = ['la ', 'cette', 'une ', 'ma ', 'ta ', 'sa ', 'chère ', "l' ancienne ", 'quelle ', 'chère', 'petite', 'première', 'seconde', 'dernière', 'belle', 'ancienne', 'malheureuse', 'grande', 'innocente', 'intéressante', 'nouvelle', 'seule', 'excellente', 'bonne']
    female_consonant_prefix += [prefix + ' ' + adj + ' ' for prefix in ['notre', 'votre', 'leur', 'pauvre', "l'", 'quelle'] for adj in ['chère', 'petite', 'première', 'seconde', 'dernière', 'belle', 'ancienne', 'malheureuse', 'grande', 'innocente', 'intéressante', 'nouvelle', 'seule', 'excellente', 'bonne']]
    female_rows = filtered_df[(filtered_df['mention_len'] <= 4)
                            &(filtered_df['mention_len'] > 1)
                            & (filtered_df['number'].isin(['Singular']))
                            & (filtered_df['gender'].isin(['Not_Assigned']))
                            & (filtered_df['text'].str.startswith(tuple(female_consonant_prefix)))
                            & (filtered_df['head_word'].isin(consonant_singular_ambiguous_head))]
    filtered_df.loc[female_rows.index, 'gender'] = 'Female'

    ambiguous_rows = filtered_df[(filtered_df['mention_len'] <= 4)
                            & (filtered_df['gender'].isin(['Not_Assigned']))
                            & (filtered_df['number'].isin(['Singular']))
                            & (filtered_df['head_word'].isin(singular_ambiguous_head))]
    filtered_df.loc[ambiguous_rows.index, 'gender'] = 'Ambiguous'

    del df
    return filtered_df
#%%
def assign_gender_to_plural_head(df):
    filtered_df = df
    filtered_df = filtered_df[filtered_df['number'] == 'Plural']
    filtered_df = filtered_df[~(filtered_df['text'].str.contains(' et '))]
    filtered_df = filtered_df[(filtered_df['mention_len'] <= 4)]

    always_plural_female_head = ['femmes', 'dames', 'filles', 'sœurs', 'amies', 'mères', 'demoiselles', 'compagnes', 'elles', 'religieuses', 'servantes', 'unes', 'maîtresses', 'princesses', 'fillettes', 'religieuses', 'petites', 'cousines', 'ouvrières', 'vieilles', 'voisines', 'danseuses', 'malheureuses', 'voyageuses', 'visiteuses', 'vendeuses', 'serveuses', 'promeneuses', 'ouvreuses', 'laveuses', 'blanchisseuses', 'chanteuses', 'voleuses', 'baigneuses', 'amoureuses', 'faneuses', 'curieuses', 'entraîneuses', 'brodeuses', 'travailleuses', 'pleureuses', 'gueuses', 'tricoteuses', 'ténébreuses', 'dormeuses', 'joueuses', 'acheteuses', 'revendeuses', 'vendangeuses', 'européennes', 'lycéennes', 'méditerranéennes', 'chicoréennes', 'vendéennes', 'dahoméennes', 'fées', 'prostituées', 'poupées', 'infortunées', 'aînées', 'protégées', 'invitées', 'condamnées', 'fiancées', 'mariées', 'employées', 'pygmées', 'épousées', 'habituées', 'âgées', 'accouchées', 'athées', 'privilégiées', 'machabées', 'subordonnées', 'intéressées', 'aimées', 'tantes', 'suivantes', 'arrivantes', 'gouvernantes', 'surveillantes', 'passantes', 'élégantes', 'habitantes', 'vivantes', 'amantes', 'aspirantes', 'étudiantes', 'jumelles']

    plural_female_rows = filtered_df[(filtered_df['head_word'].isin(always_plural_female_head))]
    df.loc[plural_female_rows.index, 'gender'] = 'Female'

    always_plural_male_head = ['garçons', 'frères', 'gars', 'fils', 'prêtres', 'moines', 'pères', 'mecs', 'gentilshommes', 'maris', 'messieurs', 'oncles']
    plural_male_rows = filtered_df[(filtered_df['head_word'].isin(always_plural_male_head))]
    df.loc[plural_male_rows.index, 'gender'] = 'Male'


    always_plural_ambiguous_head = ['anglais', 'masse', 'français', 'élèves', 'cavaliers', 'témoins', 'malheureux', 'spectateurs', 'visiteurs', 'bourgeois', 'derniers', 'espagnols', 'confrères', 'êtres', 'sentinelles', 'époux', 'adversaires', 'blancs', 'étudiants', 'jurés', 'personnages','ceux', 'deux', 'rois', 'petits', 'rois', 'vieux', 'premiers', 'riches', 'fugitifs', 'complices', 'curieux', 'quatre', 'anciens', 'jeunes', 'artistes', 'arabes', 'lecteurs', 'marins', 'princes', 'juifs', 'passagers', 'vieillards', 'dieux', 'ancêtres', 'eux', 'anges', 'magistrats', 'parisiens', 'misérables', 'chefs', 'américains', 'coupables', 'amoureux', 'journalistes', 'sauvages', 'chrétiens', 'policiers', 'colons', 'ministres', 'romains', 'employés', 'pêcheurs', 'musiciens', 'assassins', 'loups', 'amants', 'nôtre', 'brigands', 'turcs', 'valets', 'fidèles', 'chinois', 'marchands', 'serviteurs', 'sujets', 'citoyens', 'copains', 'promeneurs', 'troupes', 'poètes', 'assaillants', 'semblables', 'gardiens', 'grecs', 'joueurs', 'beaucoup', 'victimes', 'acteurs', 'touristes', 'républicains', 'cousins', 'familles', 'groupes', 'indigènes', 'savants', 'animaux', 'condamnés', 'grands', 'auditeurs', 'gosses', 'venus', 'gamins', 'esclaves', 'noirs', 'combattants', 'esclaves', 'militaires', 'nobles', 'guerriers', 'inspecteurs', 'saints', 'peintres', 'autrichiens', 'porteurs', 'petites', 'auteurs', 'courtisans', 'pompiers', 'saints', 'barbares', 'européens', 'nègres', 'adultes', 'seigneurs', 'insurgés', 'archers', 'inconnus', 'royalistes', 'types', 'peuples', 'pirates', 'oiseaux', 'bretons', 'rebelles', 'poissons', 'coquins', 'italiens', 'neveux', 'jésuites', 'assiégés', 'généraux', 'danseurs', 'aïeux', 'fous', 'conjurés', 'trois', 'travailleurs', 'locataires', 'bourreaux', 'fous', 'dragons', 'suisses', 'députés', 'professeurs', 'desquels', '-là', 'patriotes', 'avocats', 'chevaliers', 'vainqueurs', 'couples', 'propriétaires', 'contemporains', 'arrivants', 'écoliers', 'concitoyens', 'fiancés', 'partisans', 'fuyards', 'forçats', 'armées', 'couvées', 'bourgeoisie']
    plural_ambiguous_rows = filtered_df[(filtered_df['head_word'].isin(always_plural_ambiguous_head))]
    df.loc[plural_ambiguous_rows.index, 'gender'] = 'Ambiguous'

    return df
#%%
def assign_gender_to_singular_proper_mentions(entities_df):
    filtered_df = entities_df[entities_df['gender'] == 'Not_Assigned']
    filtered_df = filtered_df[filtered_df['prop'] == 'PROP']
    filtered_df = filtered_df[(filtered_df['number'].isin(['Singular']))]

    male_head_list = ['m.', 'monsieur', 'mossieur', 'monseigneur', 'don', 'sir', 'lord', 'maître', 'saint', 'messire', 'père', 'sieur', 'don', 'dom', 'comte', 'mr', 'ami', 'nommé', 'dénommé', 'dieu', 'marseillais', 'canadien', 'page', 'mgr', 'mister', 'mr.', 'baron', 'seigneur', 'frère', 'jésus', 'marquis', 'prince', 'celui', 'pasteur', 'papa', 'roi', 'master', 'herr', 'oncle', 'milord', 'mylord', 'bien-aimé', 'duc', 'vicomte', 'sire', 'abbé', 'cousin', 'parisien', 'président', 'señor', 'fils', 'tonton', 'maréchal', 'juif', 'bonhomme', 'sergent', 'captain', 'cheikh', 'christ']
    male_proper_mentions_rows = filtered_df[(filtered_df['head_word'].isin(male_head_list))]
    entities_df.loc[male_proper_mentions_rows.index, 'gender'] = 'Male'

    female_head_list = ['miss', 'mme', 'mlle', 'madame', 'mademoiselle', 'veuve', 'vierge', 'lady', 'dame', 'tante', 'mistress', 'mrs', 'mère', 'maman', 'nommée', 'dénommée', 'sœur', 'chère', 'signora', 'notre-dame', "mam'", 'sainte', 'marquise', 'reine', 'demoiselle', 'comtesse', 'bien-aimée', 'princesse', 'belle', 'petite', 'cousine', 'fille', 'parisienne', 'frau', 'duchesse', 'juive', 'elle-même', 'doña', 'dona']
    female_proper_mentions_rows = filtered_df[(filtered_df['head_word'].isin(female_head_list))]
    entities_df.loc[female_proper_mentions_rows.index, 'gender'] = 'Female'

    male_prefix_list = ['m.', 'monsieur', 'mossieur', 'monseigneur', 'don', 'sir', 'lord', 'maître', 'saint', 'messire', 'père', 'sieur', 'dom', 'comte', 'mr', 'ami', 'nommé', 'dénommé', 'dieu', 'marseillais', 'canadien', 'page', 'mgr', 'mister', 'mr.', 'baron', 'seigneur', 'frère', 'jésus', 'marquis', 'prince', 'celui', 'pasteur', 'papa', 'roi', 'master', 'herr', 'oncle', 'milord', 'mylord', 'bien-aimé', 'duc', 'cheikh', 'vicomte', 'sire', 'abbé', 'cousin', 'parisien', 'président', 'señor', 'fils', 'tonton', 'maréchal', 'juif', 'bonhomme', 'le petit', 'le vieux', 'mon petit', 'mon cher', 'le beau', 'le chanoine', 'le gros', 'mon pauvre', 'le pauvre', 'le grand', 'le captain', 'le dr', "l' évêque", 'le bel' , 'le bon', 'le cheikh', 'le jeune', 'le vieil', 'du petit', 'notre cher', 'votre cher', 'ce digne']
    male_prefix_list = [prefix + ' ' for prefix in male_prefix_list]
    male_proper_mentions_rows = filtered_df[(filtered_df['text'].str.startswith(tuple((male_prefix_list))))]
    entities_df.loc[male_proper_mentions_rows.index, 'gender'] = 'Male'


    female_prefix_list = ['miss', 'mme', 'mlle', 'madame', 'mademoiselle', 'veuve', 'vierge', 'lady', 'dame', 'tante', 'mistress', 'mrs', 'mère', 'maman', 'nommée', 'dénommée', 'sœur', 'chère', 'signora', 'notre-dame', "mam'", 'sainte', 'marquise', 'reine', 'demoiselle', 'comtesse', 'bien-aimée', 'princesse', 'belle', 'petite', 'cousine', 'fille', 'parisienne', 'frau', 'duchesse', 'juive', 'elle-même', 'la petite', 'la vieille', 'ma petite', 'ma chère', 'la belle', 'la mémé', 'ma pauvre', 'la pauvre', 'la grande', 'douairière', 'la belle', 'la jeune', 'la bonne', 'votre chère', 'notre chère', 'la bonne']
    female_prefix_list = [prefix + ' ' for prefix in female_prefix_list]
    female_proper_mentions_rows = filtered_df[(filtered_df['text'].str.startswith(tuple((female_prefix_list))))]
    entities_df.loc[female_proper_mentions_rows.index, 'gender'] = 'Female'

    del filtered_df
    return entities_df
#%%
def assign_gender_to_complex_proper_mentions(df):

    filtered_df = df[(df['gender'] == 'Not_Assigned')]
    filtered_df = filtered_df[(filtered_df['prop'] == 'PROP')]
    filtered_df = filtered_df[(filtered_df['number'] == 'Singular')]
    # filtered_df = filtered_df[(filtered_df['mention_len'] >= 3)]

    nobiliary_particle = ['de', "d'", 'des', 'du', 'af', 'von', 'le', 'la']

    female_singular_adj = ['adroite', 'affreuse', 'ancienne', 'audacieuse', 'avisée', 'belle', 'bien nommée', 'blonde', 'bonne', 'brillante', 'brutale', 'bruyante', 'charmante', 'chère', 'courageuse', 'curieuse', 'damnée', 'dangereuse', 'dernière', 'diablotine', 'divine', 'doucereuse', 'douce', 'délicieuse', 'dénommée', 'effrayante', 'ennuyeuse', 'enragée', 'excellente', 'fameuse', 'fausse', 'fichue', 'fière', 'fringuante', 'froide', 'galante', 'gentille', 'grande', 'grosse', 'généreuse', 'hardie', 'heureuse', 'hideuse', 'idiote', 'immortelle', 'impatiente', 'imprudente', 'infecte', 'infernale', 'infortunée', 'ingrate', 'innocente', 'inquiétante', 'insolente', 'insoucieuse', 'insurgée', 'intéressante', 'joyeuse', 'mme.', 'majestueuse', 'maladroite', 'malencontreuse', 'malheureuse', 'maudite', 'mauvaise', 'merveilleuse', 'monstrueuse', 'mlle.', 'mystérieuse', 'méchante', 'nommée', 'nonchalante', 'nouvelle', 'obligeante', 'obscure', 'odieuse', 'orgueilleuse', 'originale', 'petite', 'pieuse', 'piteuse', 'plaintive', 'polissonne', 'première', 'prestigieuse', 'prodigieuse', 'précieuse', 'prétendue', 'prétentieuse', 'puissante', 'ravissante', 'repoussante', 'resplendissante', 'répugnante', 'sacrée', 'salope', 'satanée', 'savante', 'seconde', 'singulière', 'sournoise', 'subtile', 'séduisante', 'valeureuse', 'vertueuse', 'vieille', 'vilaine', 'éloquente', 'élégante', 'éternelle', 'étincelante', 'étonnante', 'étourdissante', 'conquérante', 'folle', 'certaine', 'toute', 'bienheureuse', 'glorieuse', 'jolie', 'citoyenne', 'évanescente', 'brune', 'morte', 'mère', 'prudente', 'amie', 'entière']

    male_singular_adj = ['évanescent', 'adroit', 'affreux', 'ancien', 'audacieux', 'avisé', 'beau', 'bel', 'bien nommé', 'blond', 'bon', 'brillant', 'bienheureux', 'brutal', 'certain', 'bruyant', 'charmant', 'cher', 'courageux', 'curieux', 'damné', 'dangereux', 'dernier', 'diablotin', 'divin', 'doucereux', 'doux', 'délicieux', 'dénommé', 'effrayant', 'ennuyeux', 'enragé', 'excellent', 'fameux',  'faux', 'fichu', 'fier', 'fringuant', 'froid', 'galant', 'gentil', 'grand', 'gros', 'généreux', 'hardi', 'heureux', 'hideux', 'idiot', 'immortel', 'impatient', 'imprudent', 'infect', 'infernal', 'infortuné', 'ingrat', 'innocent', 'inquiétant', 'insolent', 'insoucieux', 'insurgé', 'intéressant', 'joyeux', 'm.', 'majestueux', 'maladroit', 'malencontreux', 'malheureux', 'maudit', 'mauvais', 'merveilleux', 'monstrueux', 'mr.', 'mystérieux', 'méchant', 'nommé', 'nonchalant', 'nouveau', 'nouvel', 'obligeant', 'obscur', 'odieux', 'orgueilleux', 'original', 'petit', 'pieux', 'piteux', 'plaintif', 'polisson', 'premier', 'prestigieux', 'prodigieux', 'précieux', 'prétendu', 'prétentieux', 'puissant', 'ravissant', 'repoussant', 'resplendissant', 'répugnant', 'sacré', 'salaud', 'salop', 'satané', 'savant', 'second', 'singulier', 'sournois', 'subtil', 'séduisant', 'valeureux', 'vertueux', 'vieil', 'vieux', 'vilain', 'éloquent', 'élégant', 'éternel', 'étincelant', 'étonnant', 'étourdissant', 'conquérant', 'fou', 'glorieux', 'citoyen', 'mort', 'père', 'prudent', 'ami', 'entier', 'singe', 'diable']

    ambiguous_singular_adj = ['farouche', 'abominable', 'absurde', 'admirable', 'adorable', 'aimable', 'austère', 'autre', 'avare', 'bizarre', 'brave', 'célèbre', 'docile', 'docte', 'déplorable', 'détestable', 'estimable', 'extraordinaire', 'exécrable', 'fade', 'flegmatique', 'frénétique', 'honnête', 'horrible', 'humble', 'hypocrite', 'ignoble', 'illustre', 'immense', 'immonde', 'impassible', 'improbable', 'incomparable', 'inepte', 'infâme', 'inqualifiable', 'insatiable', 'insupportable', 'intrépide', 'invincible', 'invisible', 'invraisemblable', 'irascible', 'irréprochable', 'jeune', 'juste', 'malhonnête', 'maussade', 'misérable', 'méprisable', 'même', 'noble', 'pauvre', 'pitoyable', 'poétique', 'pâle', 'redoutable', 'respectable', 'ridicule', 'robuste', 'rude', 'rustique', 'sage', 'sale', 'sauvage', 'sceptique', 'sensible', 'sinistre', 'sombre', 'stoïque', 'stupide', 'sublime', 'superbe', 'sympathique', 'terrible', 'timide', 'triste', 'vulgaire', 'énigmatique', 'énorme', 'épouvantable', 'étrange', 'fantasque', 'fantastique', 'ci-devant', 'sévère', 'simple', 'ci - devant', 'trop', 'moderne',]


    male_prefix = ['le', 'ce', 'cet', 'du', 'un', 'au']
    female_prefix = ['la', 'cette', 'ma', 'ta', 'sa', 'une', 'à la']
    ambiguous_prefix = ["l'", 'notre', 'votre', 'leur', 'mon', 'ton', 'son', 'pauvre', 'tout']


    female_attributes = [f'{prefix} {adj}' for prefix in female_prefix for adj in female_singular_adj + ambiguous_singular_adj ] + [f'{prefix} {adj}' for prefix in ambiguous_prefix for adj in female_singular_adj]
    female_rows = filtered_df[(filtered_df['gender'] == 'Not_Assigned')&
                            ((filtered_df['text'].str.contains('|'.join(map(re.escape, female_attributes)), regex=True))|(filtered_df['text'].str.startswith(tuple(['cette ', 'ma ', 'ta ', 'sa ', 'chère ', 'une ']))))]
    df.loc[female_rows.index, 'gender'] = 'Female'
    filtered_df.loc[female_rows.index, 'gender'] = 'Female'


    male_attributes = [f'{prefix} {adj}' for prefix in male_prefix for adj in male_singular_adj + ambiguous_singular_adj ] + [f'{prefix} {adj}' for prefix in ambiguous_prefix for adj in male_singular_adj]
    male_rows = filtered_df[(filtered_df['gender'] == 'Not_Assigned')&
                            ((filtered_df['text'].str.contains('|'.join(map(re.escape, male_attributes)), regex=True))|(filtered_df['text'].str.startswith(tuple(['un ', 'ce ', 'cet ', 'cher ']))))]
    df.loc[male_rows.index, 'gender'] = 'Male'

    # -- Assign gender based on suffix

    male_suffix = ['seul', "lui-même"]
    male_suffix = [f" {suffix}" for suffix in male_suffix]
    male_rows = filtered_df[(filtered_df['gender'] == 'Not_Assigned')
                            & (filtered_df['mention_len'] <= 4)
                            & (filtered_df['text'].str.endswith(tuple(male_suffix)))]
    df.loc[male_rows.index, 'gender'] = 'Male'

    return df
#%%
def assign_gender_to_proper_mentions_from_gendered_mentions(df):

    all_gendered_per_entities = df[(df['cat'] == 'PER')
                               &(df["number"] == "Singular")
                               &(df['gender'].isin(["Male", "Female"]))][['text', 'number', 'gender', 'mention_len']]

    to_gender_proper_singular = df[(df['cat'] == 'PER')
                                   &(df["number"] == "Singular")
                                   &(df["prop"] == "PROP")
                                   & ~(df['text'].str.contains(f' et '))
                                   &(df['gender'].isin(["Not_Assigned"]))][['text', 'head_word', 'mention_len', 'gender']]

    male_count_list = []
    female_count_list = []

    to_gender_proper_singular.drop_duplicates(inplace=True)

    for proper_text, proper_head in to_gender_proper_singular[['text', 'head_word']].values:
        if proper_head == "-":
            proper_head = proper_text
        if proper_text == "d' "+ proper_head:
            proper_head = proper_text

        # print(proper_head)
        proper_head = proper_head.replace('\\', '')
        # Create the regex pattern to match the whole word
        proper_head_boundaries = rf'(?<!-)\b{re.escape(proper_head)}\b(?!-)'
        # print(proper_head)
        proper_filtered = all_gendered_per_entities[(all_gendered_per_entities['number'] == 'Singular')
                             &(all_gendered_per_entities['text'].str.contains(proper_head_boundaries, regex=True))
                             & ~(all_gendered_per_entities['text'].str.contains(rf' et ')) &
                                                  ((~(all_gendered_per_entities['text'].str.contains(rf'de {proper_head}'))
                             & ~(all_gendered_per_entities['text'].str.contains(rf"d' {proper_head}"))
                             & ~(all_gendered_per_entities['text'].str.contains(rf'à {proper_head}')))
                                                    |
                             ((all_gendered_per_entities['text'].str.startswith(tuple(['ce diable de', 'ce singe de']))) & (all_gendered_per_entities['mention_len'] == 4)))
        ]

        male_count = len(proper_filtered[proper_filtered['gender'] == 'Male'])
        female_count = len(proper_filtered[proper_filtered['gender'] == 'Female'])

        male_count_list.append(male_count)
        female_count_list.append(female_count)

    to_gender_proper_singular['male_count'] = male_count_list
    to_gender_proper_singular['female_count'] = female_count_list
    to_gender_proper_singular['gendered_count'] = to_gender_proper_singular['male_count'] + to_gender_proper_singular['female_count']

    to_gender_proper_singular = to_gender_proper_singular[to_gender_proper_singular['gendered_count'] >= 1].copy()

    to_gender_proper_singular['male_ratio'] = to_gender_proper_singular['male_count'] / to_gender_proper_singular['gendered_count']
    to_gender_proper_singular['female_ratio'] = to_gender_proper_singular['female_count'] / to_gender_proper_singular['gendered_count']
    to_gender_proper_singular['max_ratio'] = np.where(to_gender_proper_singular['male_count'] > to_gender_proper_singular['female_count'], to_gender_proper_singular['male_ratio'], to_gender_proper_singular['female_ratio'])
    to_gender_proper_singular['predominant_gender'] = np.where(to_gender_proper_singular['male_count'] > to_gender_proper_singular['female_count'], 'Male', 'Female')
    to_gender_proper_singular['gender'] = np.where(to_gender_proper_singular['max_ratio'] == 1, to_gender_proper_singular['predominant_gender'], 'Ambiguous')

    male_mentions_list = to_gender_proper_singular[to_gender_proper_singular['gender'] == 'Male']["text"].tolist()
    female_mentions_list = to_gender_proper_singular[to_gender_proper_singular['gender'] == 'Female']["text"].tolist()
    ambiguous_mentions_list = to_gender_proper_singular[to_gender_proper_singular['gender'] == 'Ambiguous']["text"].tolist()

    df.loc[df[df['text'].isin(male_mentions_list)].index, 'gender'] = "Male"
    df.loc[df[df['text'].isin(female_mentions_list)].index, 'gender'] = "Female"
    df.loc[df[df['text'].isin(ambiguous_mentions_list)].index, 'gender'] = "Ambiguous"

    return df
#%%

def gender_proper_mentions_from_knowledge_base(df, insee_names_df):

    filtered_df = df[(df['prop'] == 'PROP')
                         & (df['number'] == 'Singular')
                         & (df['gender'] == 'Not_Assigned')].copy()

    original_index = filtered_df.index
    filtered_df = pd.merge(filtered_df, insee_names_df, left_on='head_word', right_on='name', how='left')
    filtered_df.index = original_index


    insee_inference_rows = filtered_df[~(filtered_df['name'].isna())
                               & (filtered_df['overall_count'] >= 20)
                               & (filtered_df['max_ratio'] >= 0.98)].copy()
    df.loc[insee_inference_rows.index.tolist(), 'gender'] = insee_inference_rows['predominant_gender'].tolist()

    # Knowledge based proper name inference ------------------------------------------------------
    male_proper_names_list = ['artagnan', 'chopin', 'athanagore', 'saint-julien', 'hitler', 'molière', 'balzac', 'homère', 'beethoven', 'mahomet', 'shakespeare', 'bonaparte', 'la fontaine', 'rousseau', 'lamartine', 'notre-seigneur', 'baudelaire', 'cicéron', 'descartes', 'stendhal', 'montaigne', 'proust', 'rabelais', 'allah', 'staline', 'goethe', 'gœthe', 'néron', 'malebranche', 'archimède', 'musset', 'schubert', 'nietzsche', 'rimbaud', 'lénine', 'sartre', 'diderot', 'kant', 'freud', 'pythagore', 'saint-germain', 'saint-léon', 'saint-françois', 'saint-georges', 'saint-vincent de paul', 'saint-mégrin', 'saint-preux', 'saint-james', 'saint-saëns', 'saint-antoine', 'saint-exupéry', "saint-thomas d' aquin", 'saint-andré', 'saint-potin', 'saint-john perse', 'saint-thomas', 'saint-vincent-de-paul', 'saint-jacques', 'saint-luc', 'saint-brice', 'saint-avit', 'saint-xist', 'saint-dieu', 'saint-nicolas', 'saint-elme', 'saint-roch', 'saint-maugon', 'saint-loup', 'saint-vincent', 'saint-hurugues', 'saint-sulpice', 'saint-rémy', 'saint-michel', 'saint-maurice', 'saint-victor', 'saint-charles', 'saint-augustin', 'saint-evremond', 'saint-guy', "saint-thomas-d'aquin", 'saint-pol', 'saint-amand', 'saint-arnaud', 'saint-benoît', 'saint-laurent', 'saint-lambert', 'saint-évremond', 'saint-vérace', 'saint-aulaire', 'saint-bernard', 'saint-philippe du roule', 'saint-hébert', 'saint-joseph', 'saint-léonard', 'saint-christophe', 'saint-cyran', 'saint-ignace', 'saint-éloi', 'saint-philippe', 'saint-séverin', 'saint-riveul', 'saint-françois de sales', 'rembrandt', 'zola', 'dostoïevski', 'ptolémée', 'télémaque', 'ronsard', 'polonius', 'béelzébuth', 'jéhovah', 'sardanapale', 'junius brutus', 'cassini', 'girodet', 'caracalla', 'humboldt', 'montesquieu', 'stendalh', 'balzac', 'potemkin', 'de balzac', 'vespasien', 'brahms', 'ïsaïe', 'verlaine', 'jacob'
]
    female_proper_names_list = ['sainte-maline', 'sainte-beuve', 'sainte-anne', 'sainte-clotilde', 'sainte-cécile', 'sainte-agnès', 'sainte-geneviève', 'sainte-marthe', "sainte-anne d' auray", 'sainte-hildegarde', 'sainte-sophie', 'sainte-brigitte', 'sainte-vierge', 'sainte-thérèse', 'sainte-périne', 'sainte-marguerite', 'sainte-odile', "sainte-claire d' ennery", 'sainte-sévère', 'sainte-aldegonde', 'sainte-austreberthe', 'sainte-foix', 'sainte-gudule', 'sainte-hermangilde', 'sainte-hermine', 'sainte-jeanne', 'sainte-misère', 'sainte-pélagie', 'sainte-justice', 'sainte-marie-des-anges', 'sainte-odile', 'sainte-orberose', 'sainte-reine', 'sainte-isabelle', 'sainte-claire-deville', 'sainte-helene', 'sainte-cécile', 'sainte-engence', 'sainte-suzanne', 'sainte-camille', 'sainte-agathe', 'sainte-amalberge', 'sainte-table', 'sainte-sophie la gigantesque', 'sainte-sidonie', 'sainte-sabine', 'sainte-routine', 'sainte-roure', 'sainte-rosalie', 'sainte-pétronille', 'sainte-ouverte', 'sainte-opportune', 'sainte-thérèse de lisieux', 'sainte-thérése', 'saintemarie', 'sainteluce', 'sainte-élisabeth', 'sainte-vierge marie', 'sainte-victoire', 'sainte-vehme', 'sainte-valère', "sainte-ursule d' uria", 'sainte-glotilde', 'sainte-germaine', 'sainte-geneviève-des-ardents', 'sainte-foire', 'sainte-foi', 'sainte-euverte', 'sainte-elisabeth de bavière', 'sainte-elisabeth', 'sainte-colombe', 'sainte-colette', 'sainte-clodilde', 'sainte-claudine', 'sainte-guillotine', 'sainte-nitouche', "sainte-n'y-touche", 'sainte-moline', 'sainte-menehould et braux - sainte - cubière', 'sainte-mary', 'sainte-marthe des lataniers', 'sainte-marie-formose', 'cendrillon', 'penthésilée', 'pomone', 'galathée', 'jj. rousseau', 'george sand']

    male_knowledge_based_rows = filtered_df[(filtered_df['text'].isin(male_proper_names_list))]
    df.loc[male_knowledge_based_rows.index, 'gender'] = 'Male'
    female_knowledge_based_rows = filtered_df[(filtered_df['text'].isin(female_proper_names_list))]
    df.loc[female_knowledge_based_rows.index, 'gender'] = 'Female'

    return df
#%%
def assign_gender_to_PER_entities(PER_entities_df, insee_path='/home/antoine/Documents/propp_modular_development/propp_fr/src/propp_fr/data/insee_names_fr_1900_2023.csv'):
    insee_names_df = pd.read_csv(insee_path,
                                 sep='\t', quoting=csv.QUOTE_NONE, low_memory=False)

    PER_entities_df['gender'] = 'Not_Assigned'
    PER_entities_df = clean_column(PER_entities_df, 'text')
    PER_entities_df = clean_column(PER_entities_df, 'head_word')

    PER_entities_df = PER_entities_df[['prop', 'cat', 'text', 'head_word', 'number', 'gender', 'mention_len']]
    grouped_PER_entities_df = PER_entities_df.groupby(['prop', 'cat', 'text', 'head_word', 'number', 'gender', 'mention_len']).first().reset_index()

    grouped_PER_entities_df = exact_ngram_match(grouped_PER_entities_df)
    grouped_PER_entities_df = always_gender_head(grouped_PER_entities_df)
    grouped_PER_entities_df = singular_ambiguous_gender_head(grouped_PER_entities_df)
    grouped_PER_entities_df = assign_gender_to_plural_head(grouped_PER_entities_df)
    grouped_PER_entities_df = assign_gender_to_singular_proper_mentions(grouped_PER_entities_df)
    grouped_PER_entities_df = assign_gender_to_complex_proper_mentions(grouped_PER_entities_df)
    grouped_PER_entities_df = assign_gender_to_proper_mentions_from_gendered_mentions(grouped_PER_entities_df)
    grouped_PER_entities_df = gender_proper_mentions_from_knowledge_base(grouped_PER_entities_df, insee_names_df)

    original_index = PER_entities_df.index
    PER_entities_df = pd.merge(PER_entities_df.drop(columns=['gender']), grouped_PER_entities_df, on=['prop', 'cat', 'text', 'head_word', 'number', 'mention_len'], how='left')
    PER_entities_df.index = original_index
    return PER_entities_df




import os
import re
import pandas as pd
import numpy as np
from tqdm.auto import tqdm

from .propp_fr_load_save_functions import load_sacr_file, save_text_file, save_tokens_df, save_entities_df
from .propp_fr_generate_tokens_df import load_spacy_model, generate_tokens_df
from .propp_fr_add_entities_features import add_features_to_entities


def clean_sacr_content(sacr_content):
    """Removes color and tokenization metadata from the end of SACR files."""
    color_idx = sacr_content.find("#COLOR")
    tokenization_idx = sacr_content.find("#TOKENIZATION-TYPE")

    # Find the first occurring metadata marker
    first_metadata_idx = min(filter(lambda x: x != -1, [color_idx, tokenization_idx]), default=-1)

    if first_metadata_idx != -1:
        sacr_content = sacr_content[:first_metadata_idx].rstrip()

    sacr_content = sacr_content.strip()
    sacr_content = re.sub(r'�', ' ', sacr_content)
    sacr_content = re.sub(r'■', ' ', sacr_content)
    sacr_content = re.sub(r'•', ' ', sacr_content)
    sacr_content = sacr_content.replace("’", "'")
    sacr_content = sacr_content.replace("' ", "'")
    sacr_content = sacr_content.replace(' .', '.')
    sacr_content = sacr_content.replace(" , ", ", ")
    sacr_content = sacr_content.replace("\xa0", " ")
    # Replace multiple spaces (but not newlines) with a single space
    sacr_content = re.sub(r'(?<=\S) {2,}(?=\S)', ' ', sacr_content)
    sacr_content = re.sub(r'[–—―‒]', '-', sacr_content)
    sacr_content = sacr_content.replace(".-", ". -")
    sacr_content = sacr_content.replace("!-", "! -")
    sacr_content = sacr_content.replace("?-", "? -")
    return sacr_content

def remove_sacr_annotations(sacr_content):
    # Remove all substrings matching the mention_oppening_pattern
    mention_oppening_pattern = r'\{[A-Za-z0-9_-]+:EN="([^"]*)"+ '
    raw_text = re.sub(mention_oppening_pattern, "", sacr_content)
    # Remove all '}' mention_closing characters
    raw_text = raw_text.replace('{', '')
    raw_text = raw_text.replace('}', '')
    raw_text = raw_text.replace("\xa0", " ")
    # Replace multiple spaces (but not newlines) with a single space
    raw_text = re.sub(r'(?<=\S) {2,}(?=\S)', ' ', raw_text)
    raw_text = raw_text.replace(' .', '.')
    raw_text = raw_text.replace(' , ', ', ')
    raw_text = raw_text.replace("’ ", "'")
    return raw_text

def get_mention_text_from_ids(start_id, end_id, text):
    return text[start_id: end_id]

def extract_entities_annotations(sacr_content):
    mention_opening_pattern = r'\{[A-Za-z0-9_-]+:EN="([^"]*)"+ '
    # Find all matches with their start and end positions
    matches = [(m.start(), m.end(), m.group()) for m in re.finditer(mention_opening_pattern, sacr_content)]
    opening_ids = sorted([start for start, end, match in matches])
    closing_ids = [i for i, char in enumerate(sacr_content) if char == "}"] # Find indices of all "}" characters

    # annotation pairs
    ordered_annotations_boundaries = []
    for annotation_opening in opening_ids:
        closing_candidates = [end for end in closing_ids if end > annotation_opening]
        for closing_candidate in closing_candidates:
            contained_annotations_opening = [start for start in opening_ids if annotation_opening < start < closing_candidate]
            contained_annotations_closing = [end for end in closing_candidates if annotation_opening < end < closing_candidate]
            if len(contained_annotations_opening) == len(contained_annotations_closing):
                ordered_annotations_boundaries.append({"sacr_start_id": annotation_opening,
                                                       "sacr_end_id": closing_candidate})
                break

    df = pd.DataFrame(ordered_annotations_boundaries)
    # Apply the function to create a new 'text' column
    df["annotation"] = df.apply(lambda row: get_mention_text_from_ids(row["sacr_start_id"], row["sacr_end_id"]+1, sacr_content), axis=1)
    # Apply regex to extract the substring between { and :EN="
    df["COREF_name"] = df["annotation"].str.extract(r'\{([A-Za-z0-9_-]+):EN="')
    # Apply regex to extract text between the first two quotation marks
    df["cat"] = df["annotation"].str.extract(r'="([^"]*)"')

    return df

def convert_ids_from_sacr_to_recovered(entities_df, sacr_content, recovered_text):
    all_ids = entities_df["sacr_start_id"].tolist() + entities_df["sacr_end_id"].tolist()
    sorted_ids = sorted(all_ids)

    sacr_to_recovered_index_dict = {}

    for sacr_index in sorted_ids:
        last_known_sacr_index = max(list(sacr_to_recovered_index_dict.keys()), default=None)
        if last_known_sacr_index:
            last_known_recovered_text_index = sacr_to_recovered_index_dict[last_known_sacr_index]
            sacr_to_recovered_delta = len(remove_sacr_annotations(sacr_content[last_known_sacr_index:sacr_index]))
            recovered_text_index = last_known_recovered_text_index + sacr_to_recovered_delta
        else:
            recovered_text_index = len(remove_sacr_annotations(sacr_content[:sacr_index]))

        sacr_to_recovered_index_dict[sacr_index] = recovered_text_index

    entities_df["byte_onset"] = entities_df["sacr_start_id"].map(sacr_to_recovered_index_dict)
    entities_df["byte_offset"] = entities_df["sacr_end_id"].map(sacr_to_recovered_index_dict)

    # Apply the function to create a new 'text' column
    entities_df["sacr_text"] = entities_df.apply(lambda row: get_mention_text_from_ids(row["byte_onset"], row["byte_offset"], recovered_text), axis=1)

    return entities_df

def get_tokens_start_end(entities_df, tokens_df):
    # Convert tokens byte onset and offset into NumPy arrays for efficient processing
    token_onsets = tokens_df['byte_onset'].values
    token_offsets = tokens_df['byte_offset'].values

    # Precompute masks for each entity
    start_tokens = []
    end_tokens = []

    for byte_onset, byte_offset in entities_df[["byte_onset", "byte_offset"]].values:
        # Efficiently find token range that overlaps with entity using NumPy boolean indexing
        start_mask = token_offsets > byte_onset
        end_mask = token_onsets < byte_offset

        # The tokens that satisfy both conditions are the ones that are part of the entity
        relevant_tokens = np.where(start_mask & end_mask)[0]

        # Get the first and last token
        start_tokens.append(relevant_tokens[0] if len(relevant_tokens) > 0 else -1)  # -1 if no token found
        end_tokens.append(relevant_tokens[-1] if len(relevant_tokens) > 0 else -1)

    # Add the results back to the entities DataFrame
    entities_df["start_token"] = start_tokens
    entities_df["end_token"] = end_tokens

    return entities_df

def reorder_coref_ids(entities_df):
    COREF_column = 'COREF_name'

    # Get the most frequent category for each COREF_name
    coref_counts = entities_df.groupby(COREF_column)['cat'].agg(lambda x: x.value_counts().idxmax())

    # Get the count of mentions per COREF_name
    coref_sizes = entities_df[COREF_column].value_counts()

    # Combine the counts and most frequent categories
    grouped_entities_df = pd.DataFrame({
        'Count': coref_sizes,
        'coref_cat': coref_counts
    }).reset_index()

    # Sort by 'Count' to get the coref_name with the most mentions first
    grouped_entities_df.sort_values(by='Count', ascending=False, inplace=True)

    # Use pd.factorize to generate new coref ids
    grouped_entities_df['new_COREF'] = pd.factorize(grouped_entities_df['COREF_name'])[0]

    # Map the old COREF_name to the new COREF id
    COREF_converter = dict(zip(grouped_entities_df[COREF_column], grouped_entities_df['new_COREF']))

    # Assign the new COREF ids back to the entities dataframe
    entities_df['COREF'] = entities_df[COREF_column].map(COREF_converter)

    return entities_df

def extract_text_for_entities(tokens_df, entities_df, recovered_text):
    # Precompute the byte onset and offset for each token
    tokens_byte_onsets = tokens_df["byte_onset"].values
    tokens_byte_offsets = tokens_df["byte_offset"].values

    # Initialize a list to store the extracted texts
    texts = []

    # Iterate over each entity's start and end token indices
    for start_token, end_token in entities_df[["start_token", "end_token"]].values:
        # Find the start and end byte offsets
        byte_onset = tokens_byte_onsets[start_token]
        byte_offset = tokens_byte_offsets[end_token]

        # Slice the text from the recovered_text using the precomputed offsets
        texts.append(recovered_text[byte_onset: byte_offset])

    # Assign the extracted texts to the DataFrame
    entities_df["text"] = texts
    return entities_df

def generate_tokens_and_entities_from_sacr(file_name,
                                           files_directory,
                                           end_directory=None,
                                           spacy_model=None,
                                           max_char_sentence_length=75000,
                                           cat_replace_dict=None,
                                           entity_types=None):
    # print(SACR_file_name)
    if cat_replace_dict is None:
        cat_replace_dict = {"f FAC": "FAC",
                            "g GPE": "GPE",
                            "h HIST": "TIME",
                            "l LOC": "LOC",
                            "m METALEPSE": "PER",
                            "n NO_PER": "PER",
                            "o ORG": "ORG",
                            "p PER": "PER",
                            "t TIME": "TIME",
                            "v VEH": "VEH",
                            "": "PER",
                            }
    if spacy_model == None:
        spacy_model = load_spacy_model(model_name='fr_dep_news_trf', model_max_length=500000)

    if end_directory==None:
        end_directory = files_directory

    sacr_file_path = os.path.join(files_directory, file_name)
    if not sacr_file_path.endswith(".sacr"):
        sacr_file_path = sacr_file_path + ".sacr"

    with open(sacr_file_path, "r", encoding="UTF-8") as sacr_file:
        sacr_content = sacr_file.read()
    sacr_content = clean_sacr_content(sacr_content)
    recovered_text = remove_sacr_annotations(sacr_content)
    entities_df = extract_entities_annotations(sacr_content)
    entities_df = convert_ids_from_sacr_to_recovered(entities_df, sacr_content, recovered_text)
    entities_df["cat"] = entities_df["cat"].map(cat_replace_dict)


    tokens_df = generate_tokens_df(recovered_text, spacy_model, max_char_sentence_length=max_char_sentence_length)

    entities_df = get_tokens_start_end(entities_df, tokens_df)

    entities_df = reorder_coref_ids(entities_df)
    # # print(entities_df['cat'].unique())
    entities_df = entities_df[['COREF_name', 'COREF', 'start_token', 'end_token', 'cat', 'sacr_text', 'byte_onset', 'byte_offset']]

    entities_df = extract_text_for_entities(tokens_df, entities_df, recovered_text)

    if entity_types:
        entities_df = entities_df[entities_df["cat"].isin(entity_types)]
    entities_df = add_features_to_entities(entities_df, tokens_df)


    # remove tokens after the last sentence with annotated entity // allow to filter partially anotated SACR files, can continue annotations at anny given time
    last_annotated_sentence = tokens_df.loc[entities_df['end_token'].max(), 'sentence_ID']
    tokens_df = tokens_df[tokens_df['sentence_ID'] <= last_annotated_sentence]


    save_text_file(recovered_text, file_name, files_directory=end_directory, extension=".txt")
    save_tokens_df(tokens_df, file_name, files_directory=end_directory, extension=".tokens")
    save_entities_df(entities_df, file_name, files_directory=end_directory, extension=".entities")





# #%%
# from .propp_fr_load_save_functions import load_sacr_file, clean_text, save_text_file, save_tokens_df, save_entities_df
# from .propp_fr_generate_tokens_df import load_spacy_model, generate_tokens_df
# from .propp_fr_add_entities_features import add_features_to_entities
# #%%
# import re
# import pandas as pd
# from tqdm.auto import tqdm
# import os
# #%%
# def remove_sacr_metadata(sacr_content):
#     #Remove color and tokenization metadata from the end of SACR files
#     sacr_metadata_index = sacr_content.find("#COLOR") if sacr_content.find("#COLOR") != -1 else sacr_content.find("#TOKENIZATION-TYPE")
#     if sacr_metadata_index != -1:
#         sacr_content = sacr_content[:sacr_metadata_index].rstrip()
#     sacr_content = sacr_content.strip()
#     return sacr_content
# def replace_entities_tags(sacr_content, cat_replace_dict):
#     for sacr_label in cat_replace_dict.keys():
#         new_label = cat_replace_dict[sacr_label]
#         sacr_content = sacr_content.replace(f':EN="{sacr_label}" ', f':EN="{new_label}" ')
#     return sacr_content
# def remove_sacr_annotations(sacr_content):
#     # Remove all substrings matching the mention_oppening_pattern
#     mention_oppening_pattern = r'\{[A-Za-z0-9_-]+:EN="([^"]*)"+ '
#     raw_text = re.sub(mention_oppening_pattern, "", sacr_content)
#     # Remove all '}' mention_closing characters
#     raw_text = raw_text.replace('}', '')
#     return raw_text
# def extract_entities(sacr_content):
#     # extracting indices of mentions span boundaries
#     opening_indices = [i for i, char in enumerate(sacr_content) if char == '{']
#
#     entities_dict = []
#
#     for opening_index in opening_indices:
#         end_index = opening_index+1
#         while len([i for i, char in enumerate(sacr_content[opening_index:end_index]) if char == '{']) != len([i for i, char in enumerate(sacr_content[opening_index:end_index]) if char == '}']):
#             end_index += 1
#
#         raw_text = sacr_content[opening_index:end_index]
#         text = remove_sacr_annotations(raw_text)
#         entity_type = re.search(r':EN="([^"]*)"', raw_text).group(1)
#         coref_name = re.search(r'{([^:}]*)', raw_text).group(1)
#
#         entities_dict.append({'SACR_start_id': opening_index,
#                               'SACR_end_id': end_index,
#                               'raw_text': raw_text,
#                               # 'text': text,
#                               'cat': entity_type,
#                               'COREF_name': coref_name})
#
#     entities_df = pd.DataFrame(entities_dict)
#     return entities_df
# def convert_annotated_ids_to_raw_ids(opening_indices, closing_indices, sacr_content):
#     raw_indices_lists = []
#     annotated_to_raw_index_dict = {}
#
#     for indices_list in [opening_indices, closing_indices]:
#         raw_ids_list = []
#         for annotated_index in indices_list:
#             max_known_index = max([v for v in annotated_to_raw_index_dict.keys() if v < annotated_index], default=None)
#             if max_known_index:
#                 raw_text_known_index = annotated_to_raw_index_dict[max_known_index]
#                 delta = len(clean_text(remove_sacr_annotations(sacr_content[max_known_index:annotated_index])))
#                 raw_index = raw_text_known_index + delta
#             else:
#                 raw_index = len(remove_sacr_annotations(sacr_content[:annotated_index]))
#
#             annotated_to_raw_index_dict[annotated_index] = raw_index
#             raw_ids_list.append(raw_index)
#
#         raw_indices_lists.append(raw_ids_list)
#
#     raw_text_opening_indices, raw_text_closing_indices = raw_indices_lists[0], raw_indices_lists[1]
#     return raw_text_opening_indices, raw_text_closing_indices
# def add_tokens_infos_to_entities(entities_df, tokens_df, sacr_content):
#     byte_onset, byte_offset = convert_annotated_ids_to_raw_ids(entities_df['SACR_start_id'], entities_df['SACR_end_id'], sacr_content)
#     entities_df['byte_onset'] = byte_onset
#     entities_df['byte_offset'] = byte_offset
#
#     start_tokens, end_tokens, text_list  = [], [], []
#     for byte_onset, byte_offset in entities_df[['byte_onset', 'byte_offset']].values:
#         sample_tokens_df = tokens_df[(tokens_df['byte_offset'] > byte_onset) & (tokens_df['byte_onset'] < byte_offset)]
#
#         token_ids = sample_tokens_df['token_ID_within_document'].tolist()
#         start_token, end_token = token_ids[0], token_ids[-1]
#         start_tokens.append(start_token)
#         end_tokens.append(end_token)
#         text_list.append(' '.join(sample_tokens_df['word'].tolist()))
#
#     entities_df['start_token'] = start_tokens
#     entities_df['end_token'] = end_tokens
#     entities_df['text'] = text_list
#
#     return entities_df
# def reorder_coref_ids(entities_df):
#     COREF_column = 'COREF_name'
#     # Group by 'COREF' column and aggregate
#     grouped_entities_df = entities_df.groupby(COREF_column).agg(
#         Count=(COREF_column, 'size'),
#         coref_cat=('cat', lambda x: x.value_counts().idxmax())  # Inline lambda function for most frequent value
#     ).reset_index()
#     # Sorting by mention count
#     grouped_entities_df = grouped_entities_df.sort_values(by=['Count'], ascending=[False])#.drop(columns=['cat_priority'])
#
#     grouped_entities_df = grouped_entities_df.reset_index(drop=True)
#     grouped_entities_df['new_COREF'] = grouped_entities_df.index
#     COREF_converter = dict(zip(grouped_entities_df[COREF_column], grouped_entities_df['new_COREF']))
#
#     entities_df['COREF'] = entities_df[COREF_column].map(COREF_converter)
#
#     return entities_df
# #%%
# def generate_tokens_and_entities_from_sacr(file_name, files_directory,
#                                                  end_directory=None,
#                                                  spacy_model=None,
#                                                  max_char_sentence_length=75000,
#                                                  cat_replace_dict=None):
#     # print(SACR_file_name)
#     if cat_replace_dict is None:
#         cat_replace_dict = {"f FAC": "FAC",
#                             "g GPE": "GPE",
#                             "h HIST": "TIME",
#                             "l LOC": "LOC",
#                             "m METALEPSE": "PER",
#                             "n NO_PER": "PER",
#                             "o ORG": "ORG",
#                             "p PER": "PER",
#                             "t TIME": "TIME",
#                             "v VEH": "VEH",
#                             "": "PER",
#                             }
#     if spacy_model == None:
#         spacy_model = load_spacy_model(model_name='fr_dep_news_trf', model_max_length=500000)
#
#     if end_directory==None:
#         end_directory = files_directory
#
#     sacr_content = load_sacr_file(file_name, files_directory=files_directory, extension=".sacr")
#     sacr_content = remove_sacr_metadata(sacr_content)
#     sacr_content = clean_text(sacr_content)
#     sacr_content = replace_entities_tags(sacr_content, cat_replace_dict)
#
#     recovered_txt_file_content = remove_sacr_annotations(sacr_content)
#     recovered_txt_file_content = clean_text(recovered_txt_file_content)
#
#     save_text_file(recovered_txt_file_content, file_name, files_directory=end_directory, extension=".txt")
#
#     tokens_df = generate_tokens_df(recovered_txt_file_content, spacy_model, max_char_sentence_length=max_char_sentence_length)
#     entities_df = extract_entities(sacr_content)
#     entities_df = add_tokens_infos_to_entities(entities_df, tokens_df, sacr_content)
#
#     entities_df = reorder_coref_ids(entities_df)
#     # # print(entities_df['cat'].unique())
#     entities_df = entities_df[['COREF_name', 'COREF', 'start_token', 'end_token', 'cat', 'text']]
#
#     entities_df = add_features_to_entities(entities_df, tokens_df)
#
#     # entities_df = entities_df[['COREF_name', 'COREF', 'start_token', 'end_token', 'cat', 'text', 'prop', 'number', 'gender', 'head_word', 'mention_len', 'head_dependency_relation', 'in_to_out_nested_level', 'out_to_in_nested_level','nested_entities_count', 'paragraph_ID', 'sentence_ID', 'start_token_ID_within_sentence', 'POS_tag', 'head_id', 'head_syntactic_head_ID']]
#     # print(Counter(entities_df['cat']))
#
#
#     # remove tokens after the last sentence with annotated entity // allow to filter partially anotated SACR files, can continue annotations at anny given time
#     last_annotated_sentence = tokens_df.loc[entities_df['end_token'].max(), 'sentence_ID']
#     tokens_df = tokens_df[tokens_df['sentence_ID'] <= last_annotated_sentence]
#
#
#
#     save_tokens_df(tokens_df, end_directory, file_name, extension=".tokens")
#     save_entities_df(entities_df, end_directory, file_name, extension=".entities")



import pandas as pd

import sys
from tqdm.auto import tqdm

import propp_fr
# Patch sys.modules to redirect references to 'propp_fr' to 'propp_fr'
sys.modules['propp_fr'] = propp_fr

from .propp_fr_load_save_functions import load_sacr_file, load_text_file, save_text_file, load_tokens_df, save_tokens_df, load_entities_df, save_entities_df, clean_text, load_book_file, save_book_file
from .propp_fr_add_entities_features import add_features_to_entities
from .propp_fr_generate_tokens_df import load_spacy_model, generate_tokens_df
from .propp_fr_generate_tokens_and_entities_from_sacr import generate_tokens_and_entities_from_sacr
from .propp_fr_generate_tokens_embeddings_tensor import load_tokenizer_and_embedding_model, get_embedding_tensor_from_tokens_df
from .propp_fr_mentions_detection_module import mentions_detection_LOOCV_full_model_training, generate_NER_model_card_from_LOOCV_directory, load_mentions_detection_model, generate_entities_df
from .propp_fr_mentions_detection_module import LockedDropout, Highway, NERModel

from .propp_fr_coreference_resolution_module import coreference_resolution_LOOCV_full_model_training, generate_coref_model_card_from_LOOCV_directory
from .propp_fr_coreference_resolution_module import load_coreference_resolution_model, perform_coreference, CoreferenceResolutionModel

from .propp_fr_extract_attributes import extract_attributes
from .propp_fr_generate_characters_dict import generate_characters_dict

from .propp_fr_generate_sacr_file import generate_sacr_file

from .propp_fr_single_line_command import process_text_file, load_models

# Inside propp_fr/__init__.py
print("propp_fr package loaded successfully.")







#%%
import pandas as pd
from collections import Counter
from tqdm.auto import tqdm
import json
import os
import csv
from time import time # To time our operations
tqdm.pandas()
#%%
def clean_column(df, column):
    df[column] = df[column].str.lower()
    df[column] = df[column].fillna('')
    # Replacement operations
    df[column] = df[column].replace(r'\!', '', regex=True)
    df[column] = df[column].replace(r'[–—―‒]', '-', regex=True)  # Replace different dash characters with hyphen
    df[column] = df[column].replace(r'\.\.\.', ' ', regex=True)
    df[column] = df[column].replace(r'\*', ' ', regex=True)
    df[column] = df[column].replace(r'"', '', regex=True)
    df[column] = df[column].replace(r'‶', '', regex=True)
    df[column] = df[column].replace(r'«', '', regex=True)
    df[column] = df[column].replace(r'\.\-', '', regex=True)  # Remove all occurrences of '.-'
    df[column] = df[column].replace(r'^[-.]', '', regex=True)  # Remove leading '-' or '.'
    df[column] = df[column].replace(r'�', '', regex=True)
    df[column] = df[column].replace(r'’', "'", regex=True)
    df[column] = df[column].replace(r'\[', " ", regex=True)
    df[column] = df[column].replace(r'\]', " ", regex=True)
    df[column] = df[column].replace(r'\(', " ", regex=True)
    df[column] = df[column].replace(r'\)', " ", regex=True)
    df[column] = df[column].replace(r'[ \t]{2,}', " ", regex=True)  # Collapse multiple spaces/tabs to a single space
    df[column] = df[column].str.strip()  # Remove leading and trailing spaces
    return df
#%%
def exact_ngram_match(df):
    filtered_df = df[df['number'] == 'Not_Assigned']
    
    singular_ngram_list = ['il', '-il', 'je', '-je', 'elle', '-elle','son', 'lui', 'sa', 'ses', 'j’', "j'", 'mon', 'tu', 'me', 'moi', 'ma', 'm’', "m'", 'le', 'l’', "l'" ,'la', 'mes', 'tes', '-il', 'te', 'toi', 'monsieur', 't’', 'ton', 'madame', 'ta', '-moi', 'maman', 'mademoiselle', 'celui-ci', 'celle-ci', "t'", 'lui-même', 'monseigneur', 'ce', 'celui', 'celle', 'cette', 'cet', 'papa', 'lequel', 'laquelle', 'quelqu’ un', 'elle-même', 'père', 'mère', 'sire', 'mienne', 'mien', 'dame', 'docteur', 'capitaine', 'seigneur', 'sien', 'homme', 'sienne', 'patron', 'femme', 'j', 'général', 'chérie', 'une seule et même personne', 'son seigneur et maître', 'une femme jeune et belle', 'auquel', "quelqu' un", 'chevalier', 'grand-père', 'commissaire', 'reine', 'prince', 'enfant', 'frère', 'petite', 'grand-mère', 'ami', 'comte', 'chéri', 'garçon', 'commandant', 'messire', 'prêtre', 'petit', 'miss', 'duc', 'chef', 'citoyen', 'comtesse', 'princesse', 'baron', 'colonel', 'mec', 'bijou', 'camarade', 'cousin', 'lieutenant', 'misérable', 'mignonne', 'maître', 'maîtresse', 'nana', 'un', 'une', 'notre-dame', 'fillette', 'duquel', 'bonhomme', 'médecin', 'cher', 'sir', 'chère', 'quelle', 'brigadier', 'parrain', 'sergent', "grand'mère", 'compère', 'professeur', 'altesse', 'vicomte', 'roi', 'berger', 'duchesse', 'cadet', 'moine', 'maréchal', 'inspecteur', 'tante', 'juge', 'tonton', 'nourrice', 'excellence', 'cousine', 'citoyenne', 'curé', 'imbécile', 'bonne-maman', 'voleur', 'tienne', 'tien', 'nounou', 'major', 'cardinal', 'mari', 'gamin', 'enfant', 'marraine', 'fiston', 'magistrat', 'soldat', 'demoiselle', 'vieillard', 'nénette', 'm’ sieur', 'monte - cristo', 'myrtille', 'reinette', 'coco', 'forestier', 'chouchou', 'marin', 'compagnon', 'brigand', 'petite-reine', 'coquin', 'assassin', 'baronne', 'gouverneur', 'canaille', 'flic', 'président', 'mémé', 'député', 'chien', 'bourreau', 'oncle', 'malheureuse', 'malin', 'amie', 'ministre', 'voisin', 'cavalier', 'mamzelle', 'saint', 'caporal', 'animal', 'fille', 'doc', 'évêque', 'bébé', 'sœur', 'neveu', 'môme', 'officier', 'trésor', 'gendre', 'avocat', 'cocher', 'démon', 'pauvre', 'mouton', 'paysan', 'maire', 'bandit', 'étranger', 'connétable', 'bouledogue', 'musicien', 'beau-père', 'grand-papa', 'gendarme', 'peste', 'parisien', 'malade', 'fidèle', 'guide', '-tu', "l'enfant", "quelqu'un", 'mouton', "celui-là", 'celle-là', 'brigand', "-j'", "- la", 'toute seule'
                          ]
    plural_ngram_list = ['nous', 'ils', 'elles', 'eux', 'leur', 'leurs', 'les', 'notre', 'nos', 'tous', 'toutes', 'ceux', 'lesquels', 'lesquelles', 'auxquels', 'auxquelles', 'hommes', 'femmes', 'mesdames', 'enfants', 'messieurs',
                        '-ils', '-nous', 'amis', 'amies', 'nôtre', 'deux', 'soldats', 'mesdemoiselles', 'citoyens', 'camarades', 'frères', 'beaucoup', 'brigands', 'seigneurs', 'prêtres', 'quelques-uns', 'celles', 'messeigneurs', 'garçons',
                         'desquels', 'nôtres', 'plusieurs', 'voyageurs', 'dames', 'parents', 'certains', 'moines', 'bandits', 'trois', 'matelots', 'treize', 'fous', 'voleurs', 'musiciens', 'marins', 'compagnons', 'cousins',
                         'maîtres', 'juges', 'médecins', 'princes', 'ministres', 'chevaliers', 'vieillards', 'courtisans', 'messires', 'valets', 'd’ autres', 'vous autres', 'vous deux', 'nous deux', 'eux deux'
                        ]
    ambiguous_ngram_list = ['vous', '-vous', 'votre', 'vôtre', 'vos', 'qui', 'que', "qu'", 'qu’','on', 'dont', "c'", 'c’', 'personne', 'chacun', 'chaque', 'en', 's’', "s'", 'bourgeois', 'malheureux', 'fils', 'gars', 'vieux', 'ça',
                            'français', 'aucun', 'aucune']

    singular_rows = filtered_df[(filtered_df['text'].isin(singular_ngram_list))]
    df.loc[singular_rows.index, 'number'] = 'Singular'
    plural_rows = filtered_df[(filtered_df['text'].isin(plural_ngram_list))]
    df.loc[plural_rows.index, 'number'] = 'Plural'
    ambiguous_rows = filtered_df[(filtered_df['text'].isin(ambiguous_ngram_list))]
    df.loc[ambiguous_rows.index, 'number'] = 'Ambiguous'
    
    del filtered_df
    return df

def et_ou_mentions(df):
    filtered_df = df[df['number'] == 'Not_Assigned']
    
    pattern = '|'.join([' et ', ' ou ', ' ni ', ' et à '])
    plural_rows = filtered_df[(filtered_df['text'].str.contains(pattern))
                              # &(filtered_df['mention_len'] <= 6)
                              &(filtered_df['in_to_out_nested_level'] > 0)
                              &(filtered_df['nested_entities_count'] >= 1)]
    df.loc[plural_rows.index, 'number'] = 'Plural'
    
    plural_rows = filtered_df[(filtered_df['text'].str.contains(pattern))
                          &(filtered_df['mention_len'] == 3)
                          &(filtered_df['in_to_out_nested_level'] > 0)]
    df.loc[plural_rows.index, 'number'] = 'Plural'

    del filtered_df
    return df

def les_mentions(df):
    filtered_df = df[df['number'] == 'Not_Assigned']

    les_rows = filtered_df[(filtered_df['text'].str.startswith(tuple(['les ', 'ces '])))]
    df.loc[les_rows.index, 'number'] = 'Plural'

    del filtered_df
    return df

def grouped_individuals(df):
    filtered_df = df[df['number'] == 'Not_Assigned']
    
    grouped_individuals_head = ['bourgeoisie', 'masse', 'tas', 'maison', 'chœur', 'peuple', 'troupeau', 'tablée', 'équipage', 'assistance', 'plupart', 'multitude', 'personnel', 'entourage', 'populace', 'orchestre', 'escorte', 'couvée', 'jeunesse', 'compagnie', 'patrouille', 'bande', 'assemblée', 'foule', 'groupe', 'famille', 'belle-famille', 'monde', 'population', 'troupe', 'noblesse', 'village', 'gens', 'tribu', 'nation', 'ville', 'société', 'colonie', 'pays', 'cité', 'régiment', 'bataillon', 'escadron', 'royaume', 'armée', 'milice', 'cohorte', 'caravane', 'horde', 'garnison', 'division', 'contrée', 'province', 'capitale', 'quartier', 'secteur', 'ordre', 'confrérie', 'communauté', 'cour', 'population', 'couple', 'police', 'ménage', 'brigade', 'public', 'gouvernement', 'auditoire', 'cortège', 'section', 'camp', 'clergé', 'paroisse', 'cavalerie', 'république', 'département', 'commune', 'corporation', 'bourg', 'comté', 'monastère', 'couvent', 'diocèse', 'ministère', 'bateau', 'conseil', 'nombreux', 'nombreuses', 'clientèle', 'personnes', 'vamile']
    plural_head = ['amis', 'diables', 'soldats', 'filles', 'femmes', 'enfants', 'hommes', 'arrivants', 'garçons', 'dames', 'connaissances', 'personnes', 'personnages', 'camarades', 'âmes', 'domestiques', 'seigneurs', 'serviteurs', 'recrues', 'parents', 'paysans', 'hôtes', 'mortels', 'messieurs', 'rentiers', 'officiers', 'coupables', 'couples', 'policiers', 'innocents', 'employés', 'chrétiens', 'témoins', 'visiteurs', 'passants', 'troupes', 'mariés', 'promeneurs', 'drôles', 'invités', 'ouvriers', 'groupes', 'êtres', 'gardes', 'marchands', 'misérables', 'spectateurs', 'combattants', 'équipages', 'gendarmes', 'journalistes', 'mères', 'autres', 'habitants', 'messieurs', 'messieurs', 'mesdames', 'mlles', 'mmes', 'mm.', 'mesdemoiselles']
    grouped_individuals_rows = filtered_df[(filtered_df['head_word'].isin(grouped_individuals_head+plural_head))]
    df.loc[grouped_individuals_rows.index, 'number'] = 'Plural'
    del filtered_df
    return df

def proper_mention(df):
    filtered_df = df[df['number'] == 'Not_Assigned']
    filtered_df = filtered_df[filtered_df['prop'] == 'PROP']
    
    ambiguous_proper_mentions = filtered_df[(filtered_df['text'].str.startswith(tuple(['des '])))
                                      # &(filtered_df['mention_len'] <= 4)
                                        ]
    df.loc[ambiguous_proper_mentions.index, 'number'] = 'Ambiguous'
    
    singular_proper_mention_rows = filtered_df[~(filtered_df['text'].str.startswith(tuple(['des '])))
                                                 &(filtered_df['number'] == 'Not_Assigned')
                                               # &(filtered_df['mention_len'] <= 4)
                                        ]
    df.loc[singular_proper_mention_rows.index, 'number'] = 'Singular'
    del filtered_df
    return df

def prefix_match(df):
    filtered_df = df[df['number'] == 'Not_Assigned']
    
    singular_prefix_list = [prefix+' ' for prefix in ['au', 'baron', 'beau', 'bel', 'belle', 'bon', 'bonne', 'ce', 'celle', 'celui', 'cet', 'cette', 'cher', 'chère', 'comte', 'cousin', 'cousine', 'dame', 'docteur', 'dom', 'don', 'donna', 'don', 'dom'
                                                      'dr', 'du', 'elle', 'feu', 'jeune', "l'", "l' un", "l' une", 'la', 'lady', 'le', 'leur', 'lord', 'lui', 'l’', 'l’ un', 'l’ une', 'm.', 'ma', 'madame', 'mademoiselle',
                                                      'maman', 'marquise', 'maître', 'me', 'messire', 'mgr', 'milord', 'miss', 'mister', 'mistress', 'mlle', 'mlle.', 'mme', 'mme.', 'moi', 'mon', 'monseigneur', 'monsieur', 'mr',
                                                      'mr.', 'mrs', 'mère', 'notre', 'papa', 'pauvre', 'petit', 'petite', 'père', 'grande', 'quel', 'quelle', 'sa', 'saint', 'sainte', 'seigneur', 'sir', 'sir.', 'soi', 'son', 'sœur', 'ta',
                                                      'tante', 'toi', 'ton', 'un', 'une', 'votre', 'tonton', 'brave', 'malheureuse', 'oncle', 'grand', 'honnête']] + ["l’", "l'"]
    plural_prefix_list = [prefix+' ' for prefix in ['autres', 'aux', 'beaucoup', 'celles', 'cent', 'certaines', 'certains', 'ces', 'ceux', 'chers', 'chères', 'bonnes', 'cinq', 'cinquante', 'dames', 'des', 'deux', 'dix', 'dons', 'douze',
                                                    'elles', 'eux', 'femmes', 'feux', 'hommes', 'huit', 'ils', 'les', 'les deux', 'leurs', 'mes', 'mesdames', 'messieurs', 'messires', 'messrs', 'mgrs', 'mille', 'misters',
                                                    'mistresses', 'mm.', 'mm .', 'mmes', 'neuf', 'nombreux', 'nos', 'nous', 'onze', 'petits', 'plusieurs', 'quarante', 'quatre', 'quelques', 'quels', 'saintes', 'saints', 'sept', 'ses',
                                                    'six', 'soixante', 'tes', 'tous', 'tout', 'toute', 'toutes', 'trente', 'trois', 'vingt', 'vos', 'quatorze']]
    ambiguous_prefix_list = [prefix+' ' for prefix in ['vous', 'aucun', 'aucune']]
    
    filtered_df = filtered_df[(filtered_df['mention_len'] > 1) & (filtered_df['mention_len'] <= 4)]
    singular_rows = filtered_df[filtered_df['text'].str.startswith(tuple(singular_prefix_list))]
    df.loc[singular_rows.index, 'number'] = 'Singular'
    plural_rows = filtered_df[filtered_df['text'].str.startswith(tuple(plural_prefix_list))]
    df.loc[plural_rows.index, 'number'] = 'Plural'
    ambiguous_rows = filtered_df[filtered_df['text'].str.startswith(tuple(ambiguous_prefix_list))]
    df.loc[ambiguous_rows.index, 'number'] = 'Ambiguous'
   
    del filtered_df
    return df
#%%
def assign_number_to_PER_entities(PER_entities_df):
    PER_entities_df['number'] = 'Not_Assigned'
    PER_entities_df = clean_column(PER_entities_df, 'text')
    PER_entities_df = clean_column(PER_entities_df, 'head_word')
    
    PER_entities_df = exact_ngram_match(PER_entities_df)
    PER_entities_df = et_ou_mentions(PER_entities_df)
    PER_entities_df = les_mentions(PER_entities_df)
    PER_entities_df = grouped_individuals(PER_entities_df)
    PER_entities_df = proper_mention(PER_entities_df)
    PER_entities_df = prefix_match(PER_entities_df)
    return PER_entities_df



import os
import spacy
import torch
import traceback
import subprocess
import sys
import json

from .propp_fr_load_save_functions import load_text_file, clean_text, save_text_file, save_tokens_df, save_entities_df, save_book_file
from .propp_fr_generate_tokens_df import load_spacy_model, generate_tokens_df
from .propp_fr_mentions_detection_module import load_mentions_detection_model, generate_entities_df
from .propp_fr_coreference_resolution_module import load_coreference_resolution_model,perform_coreference
from .propp_fr_generate_tokens_embeddings_tensor import get_embedding_tensor_from_tokens_df, load_tokenizer_and_embedding_model
from .propp_fr_add_entities_features import add_features_to_entities
from .propp_fr_extract_attributes import extract_attributes
from .propp_fr_generate_characters_dict import generate_characters_dict

def install_spacy_for_cuda():
    #print("Checking SpaCy installation...")
    
    # Determine the correct package
    cuda_version = "cpu"
    if torch.version.cuda:
        cuda_version = torch.version.cuda.split(".")[0] + "x"
    package = f"spacy[cuda{cuda_version}]" if cuda_version != "cpu" else "spacy"
    
    # Check if spacy is already installed
    try:
        import spacy
        #print(f"SpaCy {spacy.__version__} is already installed.")
        
        # Check if GPU is available and matches the desired CUDA version
        if cuda_version != "cpu":
            try:
                spacy.require_gpu()
                #print(f"SpaCy is already configured to use GPU with CUDA {cuda_version}. Skipping installation.")
                return
            except Exception:
                print("GPU support is missing or incompatible. Reinstalling...")
        else:
            print("CPU version detected. No GPU required. Skipping installation.")
            return
    except ImportError:
        print("SpaCy not found. Installing...")

    # Install or upgrade
    subprocess.run([sys.executable, "-m", "pip", "install", "--upgrade", package], check=True)
    print("SpaCy installation complete.")



def load_models(spacy_model_name="fr_dep_news_trf",
                mentions_detection_model_name="AntoineBourgois/propp-fr_NER_camembert-large_FAC_GPE_LOC_PER_TIME_VEH",
                coreference_resolution_model_name="AntoineBourgois/propp-fr_coreference-resolution_camembert-large_PER",
                spacy_input_max_length=500000,
                force_download=False): \
        # Load models once
    print("Loading models...")
    install_spacy_for_cuda()
    try:
        if torch.cuda.is_available():
            spacy.require_gpu()  # prefer
            spacy_model = spacy.load(spacy_model_name)
            print("CUDA is available, Spacy model should run on GPU.")
        else:
            spacy_model = load_spacy_model(model_name=spacy_model_name, model_max_length=spacy_input_max_length)
            print("CUDA is not available, Spacy model will run on CPU.")
    except OSError:
        subprocess.run([sys.executable, '-m', 'spacy', 'download', spacy_model_name], check=True)
        if torch.cuda.is_available():
            spacy.require_gpu()  # prefer
            spacy_model = spacy.load(spacy_model_name)
            print("CUDA is available, Spacy model should run on GPU.")
        else:
            spacy_model = load_spacy_model(model_name=spacy_model_name, model_max_length=spacy_input_max_length)
            print("CUDA is not available, Spacy model will run on CPU.")
    
    
    
    spacy_model.max_length = spacy_input_max_length

    mentions_detection_model = load_mentions_detection_model(
        model_path=mentions_detection_model_name,
        force_download=force_download
    )
    coreference_resolution_model = load_coreference_resolution_model(
        model_path=coreference_resolution_model_name,
        force_download=force_download
    )

    if mentions_detection_model["base_model_name"] != coreference_resolution_model["base_model_name"]:
        print(
            f"WARNING: mentions_detection_model and coreference_resolution_model DO NOT use the same embedding model.")
    
    print(f"\nModels Loaded Successfully:\nSpacy: {spacy_model_name}\nMentions Detection: {mentions_detection_model_name}\nCoreference Resolution: {coreference_resolution_model_name}")

    return spacy_model, mentions_detection_model, coreference_resolution_model


def process_file(file_name,
                 input_folder,
                 output_folder,
                 spacy_model,
                 mentions_detection_model,
                 coreference_resolution_model,
                 tokenizer, embedding_model,
                 spacy_max_characters_batch=1000000,
                 embedding_mini_batch=64,
                 mentions_detection_batch=14,
                 coreference_resolution_batch=50000,
                 verbose=1
                 ):
    """
    Process a single text and save outputs to the specified folder.
    """

    # Process the file
    txt_content = load_text_file(file_name, files_directory=input_folder)
    txt_content = clean_text(txt_content)
    tokens_df = generate_tokens_df(txt_content,
                                   spacy_model,
                                   max_char_sentence_length=spacy_max_characters_batch
                                   )

    # Generate token embeddings
    if verbose: print("Generating token embeddings...")
    tokens_embedding_tensor = get_embedding_tensor_from_tokens_df(txt_content,
                                                                  tokens_df,
                                                                  tokenizer,
                                                                  embedding_model,
                                                                  sliding_window_size='max',
                                                                  mini_batch_size=embedding_mini_batch,
                                                                  sliding_window_overlap=0.5,
                                                                  subword_pooling_strategy=mentions_detection_model[
                                                                      "subword_pooling_strategy"])

    if verbose: print("Mentions Detection...")
    entities_df = generate_entities_df(tokens_df,
                                       tokens_embedding_tensor,
                                       mentions_detection_model,
                                       batch_size=mentions_detection_batch
                                       )

    entities_df = add_features_to_entities(entities_df, tokens_df)

    # Perform Coreference Resolution
    if coreference_resolution_model["base_model_name"] != mentions_detection_model["base_model_name"]:
        tokenizer, embedding_model = load_tokenizer_and_embedding_model(coreference_resolution_model["base_model_name"])
        tokens_embedding_tensor = get_embedding_tensor_from_tokens_df(txt_content,
                                                                      tokens_df,
                                                                      tokenizer,
                                                                      embedding_model,
                                                                      sliding_window_size='max',
                                                                      mini_batch_size=embedding_mini_batch,
                                                                      sliding_window_overlap=0.5,
                                                                      subword_pooling_strategy=mentions_detection_model[
                                                                          "subword_pooling_strategy"])

    if verbose: print("Coreference Resolution...")
    entities_df = perform_coreference(entities_df=entities_df,
                                      tokens_embedding_tensor=tokens_embedding_tensor,
                                      coreference_resolution_model=coreference_resolution_model,
                                      batch_size=coreference_resolution_batch,
                                      propagate_coref=True,
                                      rule_based_postprocess=False)

    # Extract character attributes
    tokens_df = extract_attributes(entities_df, tokens_df)
    characters_dict = generate_characters_dict(tokens_df, entities_df)

    # Save all files
    save_text_file(txt_content, file_name, files_directory=output_folder)
    save_tokens_df(tokens_df, file_name, files_directory=output_folder)
    save_entities_df(entities_df, file_name, files_directory=output_folder)

    save_book_file(characters_dict, file_name, files_directory=output_folder)


def process_text_file(txt_file_path):
    if not os.path.isfile(txt_file_path):
        print(f"File {txt_file_path} does not exist.")

    else:
        spacy_model_name = "fr_dep_news_trf"
        spacy_input_max_length = 500000

        print("Loading models...")
        if torch.cuda.is_available():
            spacy.prefer_gpu()  # prefer
            spacy_model = spacy.load(spacy_model_name)
            print("CUDA is available, Spacy model should run on GPU.")
        else:
            spacy_model = load_spacy_model(model_name=spacy_model_name, model_max_length=spacy_input_max_length)
            print("CUDA is not available, Spacy model will run on CPU.")
        spacy_model.max_length = spacy_input_max_length

        # Split into directory and file name
        input_folder = os.path.dirname(txt_file_path)
        file_name = os.path.basename(txt_file_path)
        file_name, _ = os.path.splitext(file_name)

        output_folder = input_folder

        # Create the output folder if it doesn't exist
        os.makedirs(output_folder, exist_ok=True)

        # 1. Loading Models
        spacy_model, mentions_detection_model, coreference_resolution_model = load_models(
            spacy_model_name="fr_dep_news_trf",
            # mentions_detection_model_name = "AntoineBourgois/propp-fr_NER_camembert-large_FAC_GPE_LOC_PER_TIME_VEH",
            mentions_detection_model_name="AntoineBourgois/propp-fr_NER_camembert-large_PER",
            coreference_resolution_model_name="AntoineBourgois/propp-fr_coreference-resolution_camembert-large_PER",
            force_download=False)
        tokenizer, embedding_model = load_tokenizer_and_embedding_model(mentions_detection_model["base_model_name"])

        if os.path.exists(os.path.join(output_folder, f"{file_name}.book")):
            print(f"{file_name}.book already exists. Skipping...")

        else:
            process_file(file_name,
                         input_folder,
                         output_folder,
                         spacy_model,
                         mentions_detection_model,
                         coreference_resolution_model,
                         tokenizer, embedding_model,
                         verbose=1,
                         spacy_max_characters_batch=50000,
                         embedding_mini_batch=10,
                         mentions_detection_batch=12,
                         coreference_resolution_batch=50000,

                         # spacy_max_characters_batch=1000000,
                         # embedding_mini_batch=64,
                         # mentions_detection_batch=128,
                         # coreference_resolution_batch=50000,

                         )

            try:
                process_file(file_name,
                             input_folder,
                             output_folder,
                             spacy_model,
                             mentions_detection_model,
                             coreference_resolution_model,
                             tokenizer, embedding_model,
                             verbose=1,
                             spacy_max_characters_batch=50000,
                             embedding_mini_batch=10,
                             mentions_detection_batch=12,
                             coreference_resolution_batch=50000,

                             # spacy_max_characters_batch=1000000,
                             # embedding_mini_batch=64,
                             # mentions_detection_batch=128,
                             # coreference_resolution_batch=50000,

                             )
            except Exception as e:
                print(f"⚠️ Unexpected error processing file {file_name}: {e}")
                traceback.print_exc()  # <- prints the full traceback
